# Copyright (C) 2014-2018 Enzien Audio, Ltd.
# Copyright (C) 2023-2024 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Dict, List, Optional

from .HeavyIrObject import HeavyIrObject
from .HeavyGraph import HeavyGraph

from hvcc.types.IR import IROnMessage


class HIrOutlet(HeavyIrObject):
    """ A specific implementation of the outlet object.
    """

    def __init__(
        self,
        obj_type: str,
        args: Optional[Dict] = None,
        graph: Optional[HeavyGraph] = None,
        annotations: Optional[Dict] = None
    ) -> None:
        super().__init__("__outlet", args=args, graph=graph, annotations=annotations)

    def get_ir_on_message(self, inlet_index: int = 0) -> List[IROnMessage]:
        x = []
        if self.graph is not None:
            for c in self.graph.outlet_connections[self.args["index"]]:
                x.extend(c.to_object.get_ir_on_message(c.inlet_index))
        return x



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
# Copyright (C) 2023-2024 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from struct import unpack, pack
from typing import Callable, Dict, List, Union

from hvcc.types.IR import IROnMessage, IRSignalList, IRBuffer, IRObjectdict


class HeavyObject:

    c_struct = ""
    preamble = ""

    __C_BUFFER_DICT = {
        "~f>": "Bf",
        "~i>": "Bi",
        "input": "I",
        "output": "O",
        "zero": "ZERO"
    }

    @classmethod
    def get_C_struct(cls, obj_type: str = "") -> str:
        return cls.c_struct

    @classmethod
    def get_preamble(cls, obj_type: str) -> str:
        return cls.preamble

    @classmethod
    def get_C_header_set(self) -> set:
        return set()

    @classmethod
    def get_C_file_set(self) -> set:
        return set()

    @classmethod
    def get_C_def(cls, obj_type: str, obj_id: str) -> List[str]:
        return ["{0} {1}_{2};".format(
            cls.get_C_struct(obj_type),
            cls.get_preamble(obj_type),
            obj_id)]

    @classmethod
    def get_C_free(cls, obj_type: str, obj_id: str, args: Dict) -> List[str]:
        return ["{0}_free(&{0}_{1});".format(
            cls.get_preamble(obj_type),
            obj_id)]

    @classmethod
    def get_C_decl(cls, obj_type: str, obj_id: str, args: Dict) -> List[str]:
        return ["{0}_{1}_sendMessage(HeavyContextInterface *, int, const HvMessage *);".format(
                cls.get_preamble(obj_type),
                obj_id)]

    @classmethod
    def get_C_impl(
        cls,
        obj_type: str,
        obj_id: str,
        on_message_list: List[List[IROnMessage]],
        get_obj_class: Callable,
        objects: Dict[str, IRObjectdict]
    ) -> List[str]:
        send_message_list = [
            "{0}_{1}_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {{".format(
                cls.get_preamble(obj_type),
                obj_id)]
        if len(on_message_list) == 1:
            # if there is only one outlet, skip the switch-case template
            send_message_list.extend(
                cls._get_on_message_list(on_message_list[0], get_obj_class, objects))
        else:
            send_message_list.append("switch (letIn) {")
            for i in range(len(on_message_list)):
                send_message_list.append(f"case {i}: {{")
                send_message_list.extend(
                    cls._get_on_message_list(on_message_list[i], get_obj_class, objects))
                send_message_list.append("break;")
                send_message_list.append("}")  # end case
            send_message_list.append("default: return;")
            send_message_list.append("}")  # end switch
        send_message_list.append("}")  # end function
        return send_message_list

    @classmethod
    def get_C_init(cls, obj_type: str, obj_id: str, args: Dict) -> List[str]:
        raise NotImplementedError("method get_C_init not implemented")

    @classmethod
    def get_C_onMessage(cls, obj_type: str, obj_id: str, inlet_index: int, args: Dict) -> List[str]:
        raise NotImplementedError("method get_C_onMessage not implemented")

    @classmethod
    def get_C_process(cls, process_dict: IRSignalList, obj_type: str, obj_id: str, args: Dict) -> List[str]:
        raise NotImplementedError("method get_C_process not implemented")

    @classmethod
    def get_table_data_decl(cls, obj_type: str, obj_id: str, args: Dict) -> List[str]:
        raise NotImplementedError("method get_table_data_decl not implemented")

    @classmethod
    def _get_on_message_list(
        cls,
        on_message_list: List[IROnMessage],
        get_obj_class: Callable,
        objects: Dict[str, IRObjectdict]
    ) -> List:
        out_list = []
        for om in on_message_list:
            out_list.extend(
                get_obj_class(objects[om.id].type).get_C_onMessage(
                    objects[om.id].type,
                    om.id,
                    om.inletIndex,
                    objects[om.id].args))
        return out_list

    @classmethod
    def _c_buffer(cls, buffer_dict: IRBuffer) -> str:
        """ Returns the C represenation of the given buffer.
        """
        if buffer_dict.type == "zero":
            return cls.__C_BUFFER_DICT[buffer_dict.type]
        else:
            return "{0}{1}".format(
                cls.__C_BUFFER_DICT[buffer_dict.type],
                buffer_dict.index)

    @classmethod
    def get_hash(cls, x: Union[float, str]) -> int:
        """ Compute the message element hash used by msg_getHash().
        Returns a 32-bit integer.
        """
        if isinstance(x, float) or isinstance(x, int):
            # interpret the float bytes as an unsigned integer
            return unpack("@I", pack("@f", float(x)))[0]
        elif x == "bang":
            return 0xFFFFFFFF
        elif isinstance(x, str):
            # this hash is based MurmurHash2
            # http://en.wikipedia.org/wiki/MurmurHash
            # https://sites.google.com/site/murmurhash/
            x = str(x)
            m = 0x5bd1e995
            r = 24
            h = len(x)
            i = 0
            while i < len(x) & ~0x3:
                k = unpack("@I", bytes(x[i:i + 4], encoding='utf-8'))[0]
                k = (k * m) & 0xFFFFFFFF
                k ^= k >> r
                k = (k * m) & 0xFFFFFFFF
                h = (h * m) & 0xFFFFFFFF
                h ^= k
                i += 4

            n = len(x) & 0x3
            x = x[i:i + n]
            if n >= 3:
                h ^= (ord(x[2]) << 16) & 0xFFFFFFFF
            if n >= 2:
                h ^= (ord(x[1]) << 8) & 0xFFFFFFFF
            if n >= 1:
                h ^= ord(x[0])
                h = (h * m) & 0xFFFFFFFF

            h ^= h >> 13
            h = (h * m) & 0xFFFFFFFF
            h ^= h >> 15
            return h
        else:
            raise Exception("Message element hashes can only be computed for float and string types.")

    @classmethod
    def get_hash_string(cls, x: Union[float, str]) -> str:
        """ Returns the hash as a hex string.
        """
        return f"0x{cls.get_hash(x):X}"



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
# Copyright (C) 2023 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Optional, List, Dict

from .PdObject import PdObject
from .PdRaw import parse_pd_raw_args, PdRawException


class PdReceiveObject(PdObject):

    __INSTANCE_COUNTER = 0

    def __init__(
        self,
        obj_type: str,
        obj_args: Optional[List] = None,
        pos_x: int = 0,
        pos_y: int = 0
    ) -> None:
        assert obj_type in {"r", "r~", "receive", "receive~", "catch~"}
        super().__init__(obj_type, obj_args, pos_x, pos_y)

        self.__receiver_name = ""
        self.__extern_type = None
        self.__attributes: Dict = {}
        self.__priority = None  # priority is not set

        PdReceiveObject.__INSTANCE_COUNTER += 1
        self.__instance = PdReceiveObject.__INSTANCE_COUNTER

        try:
            # receive objects don't necessarily need to have a name
            self.__receiver_name = self.obj_args[0]

            # only extern control rate receivers
            if obj_type in {"r", "receive"}:
                # NOTE(mhroth): the second argument is _either_ externing the receiver or setting the priority
                # This means that right now priority cannot be set on externed receivers
                if self.obj_args[1] == "@hv_param":
                    self.__extern_type = "param"
                elif self.obj_args[1] == "@hv_event":
                    self.__extern_type = "event"
                elif int(self.obj_args[1]):
                    self.__priority = int(self.obj_args[1])
        except Exception:
            pass

        if self.__extern_type == "param":
            try:
                self.__attributes = {
                    "min": 0.0,
                    "max": 1.0,
                    "default": 0.5,
                    "type": "float"
                }
                self.__attributes["min"] = float(self.obj_args[2])
                self.__attributes["max"] = float(self.obj_args[3])
                self.__attributes["default"] = float(self.obj_args[4])
                self.__attributes["type"] = str(self.obj_args[5])
            except ValueError:
                self.add_warning(
                    f"Minimum, maximum, and default values for Parameter {self.__receiver_name}  must be numbers.")
            except Exception:
                pass

            if not (self.__attributes["min"] <= self.__attributes["default"]):
                self.add_error("Default parameter value is less than the minimum. "
                               "Receiver will not be exported: {0:g} < {1:g}".format(
                                   self.__attributes["default"],
                                   self.__attributes["min"]))
                self.__extern_type = None
            if not (self.__attributes["default"] <= self.__attributes["max"]):
                self.add_error("Default parameter value is greater than the maximum. "
                               "Receiver will not be exported: {0:g} > {1:g}".format(
                                   self.__attributes["default"],
                                   self.__attributes["max"]))
                self.__extern_type = None

        if '@raw' in self.obj_args or '@owl' in self.obj_args:  # TODO(dromer): deprecate @owl on next stable release
            try:
                pd_raw_args = parse_pd_raw_args(self.obj_args)
                self.__attributes.update(pd_raw_args)
                self.__extern_type = "param"  # make sure output code is generated
            except PdRawException as e:
                self.add_error(str(e))

    def validate_configuration(self) -> None:
        if self.obj_type in {"r~", "receive~"}:
            if len(self._inlet_connections.get("0", [])) > 0:
                self.add_error("[receive~] inlet connections are not supported.")

    def to_hv(self) -> Dict:
        # note: control rate send objects should not modify their name argument
        names = {
            "r": "",
            "receive": "",
            "r~": "sndrcv_sig_",
            "receive~": "sndrcv_sig_",
            "catch~": "thrwctch_sig_"
        }

        # NOTE(mhroth): we follow Pd's execution rule: deeper receivers fire first.
        # Receivers on the same level fire in the order of instantiation.
        if self.parent_graph is not None and \
                ((self.__priority is None) or (self.__receiver_name == "__hv_init" and self.__priority == 0)):
            self.__priority = (self.parent_graph.get_depth() * 1000) - self.__instance

        return {
            "type": "receive",
            "args": {
                "name": f"{names[self.obj_type]}{self.__receiver_name}",
                "extern": self.__extern_type,
                "attributes": self.__attributes,
                "priority": self.__priority
            },
            "properties": {
                "x": self.pos_x,
                "y": self.pos_y
            },
            "annotations": {
                "scope": "public"
            }
        }



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
# Copyright (C) 2023-2024 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re
from typing import Callable, Dict, List

from .HeavyObject import HeavyObject

from hvcc.types.IR import IRObjectdict, IROnMessage


class ControlMessage(HeavyObject):

    preamble = "cMsg"

    @classmethod
    def get_C_onMessage(cls, obj_type: str, obj_id: str, inlet_index: int, args: Dict) -> List[str]:
        return [f"cMsg_{obj_id}_sendMessage(_c, 0, m);"]

    @classmethod
    def get_C_impl(
        cls,
        obj_type: str,
        obj_id: str,
        on_message_list: List[List[IROnMessage]],
        get_obj_class: Callable,
        objects: Dict[str, IRObjectdict],
    ) -> List[str]:
        send_message_list = [
            f"cMsg_{obj_id}_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const n) {{"
        ]

        if len(objects[obj_id].args["local"]) > 0:
            # declare the outgoing messages (if there are any)
            send_message_list.append("HvMessage *m = nullptr;")

        # for each message
        for m in objects[obj_id].args["local"]:
            # construct the message
            send_message_list.append(f"m = HV_MESSAGE_ON_STACK({len(m)});")
            send_message_list.append(f"msg_init(m, {len(m)}, msg_getTimestamp(n));")
            for i in range(len(m)):
                e = m[i]  # get the message element
                try:
                    # is the message a float?
                    send_message_list.append(f"msg_setFloat(m, {i}, {float(e)}f);")
                except Exception:
                    if e in ["bang"]:
                        # is the message a bang?
                        send_message_list.append(f"msg_setBang(m, {i});")
                    elif re.match(r"\$[\d]+", e):
                        send_message_list.append(f"msg_setElementToFrom(m, {i}, n, {int(e[1:]) - 1});")
                    elif e == "@HV_N_SIMD":
                        # NOTE(mhroth): messages can contain special arguments
                        # which are interpreted differently than other strings
                        send_message_list.append(f"msg_setFloat(m, {i},  static_cast<float>(HV_N_SIMD));")
                    else:
                        send_message_list.append(f"msg_setSymbol(m, {i}, \"{e}\");")

            # send the message to all receiving objects
            for om in on_message_list[0]:
                send_message_list.extend(
                    get_obj_class(objects[om.id].type).get_C_onMessage(
                        objects[om.id].type,
                        om.id,
                        om.inletIndex,
                        objects[om.id].args))

        send_message_list.append("}")  # end function
        return send_message_list



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
# Copyright (C) 2023 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Optional, Dict

from .HeavyLangObject import HeavyLangObject
from .HeavyIrObject import HeavyIrObject
from .HeavyGraph import HeavyGraph


class HLangPrint(HeavyLangObject):
    """ Handles the HeavyLang "print" object.
    """

    def __init__(
        self,
        obj_type: str,
        args: Dict,
        graph: 'HeavyGraph',
        annotations: Optional[Dict] = None
    ) -> None:
        assert obj_type == "print"
        super().__init__(obj_type, args, graph,
                         num_inlets=1,
                         num_outlets=0,
                         annotations=annotations)

    def reduce(self) -> tuple:
        x = HeavyIrObject("__print", self.args)
        return ({x}, self.get_connection_move_list(x))



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
# Copyright (C) 2023-2024 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import json
import os
import time

from typing import Optional

from .HeavyException import HeavyException
from .HeavyParser import HeavyParser

from hvcc.types.compiler import CompilerResp, CompilerNotif, CompilerMsg


class hv2ir:

    @classmethod
    def compile(
        cls,
        hv_file: str,
        ir_file: str,
        patch_name: Optional[str] = None,
        verbose: bool = False
    ) -> CompilerResp:
        """ Compiles a HeavyLang file into a HeavyIR file.
            Returns a tuple of compile time in seconds, a notification dictionary,
            and a heavy object counter.
        """

        # keep track of the total compile time
        tick = time.time()

        hv_file = os.path.abspath(os.path.expanduser(hv_file))
        ir_file = os.path.abspath(os.path.expanduser(ir_file))

        try:
            # parse heavy file
            hv_graph = HeavyParser.graph_from_file(hv_file=hv_file, xname=patch_name)
        except HeavyException as e:
            return CompilerResp(
                stage="hv2ir",
                compile_time=time.time() - tick,
                notifs=CompilerNotif(
                    has_error=True,
                    exception=e,
                    errors=[CompilerMsg(message=e.message)],
                    warnings=[]
                ),
                in_file=os.path.basename(hv_file),
                in_dir=os.path.dirname(hv_file),
                out_file=os.path.basename(ir_file),
                out_dir=os.path.dirname(ir_file)
            )

        try:
            # get a counter of all heavy objects
            hv_counter = hv_graph.get_object_counter(recursive=True)

            # prepare the graph for exporting
            hv_graph.prepare()

            # ensure that the output directory exists
            if not os.path.exists(os.path.dirname(ir_file)):
                os.makedirs(os.path.dirname(ir_file))

            # generate Heavy.IR
            ir = hv_graph.to_ir()
        except HeavyException as e:
            return CompilerResp(
                stage="hv2ir",
                compile_time=time.time() - tick,
                notifs=CompilerNotif(
                    has_error=True,
                    exception=e,
                    errors=[CompilerMsg(message=e.message)],
                    warnings=[]
                ),
                in_file=os.path.basename(hv_file),
                in_dir=os.path.dirname(hv_file),
                out_file=os.path.basename(ir_file),
                out_dir=os.path.dirname(ir_file),
                obj_counter=hv_counter
            )

        if ir is not None:
            # write the hv.ir file
            with open(ir_file, "w") as f:
                if verbose:
                    json.dump(
                        ir.model_dump(),
                        f,
                        sort_keys=True,
                        indent=2,
                        separators=(",", ": "))
                else:
                    json.dump(ir.model_dump(), f)

            if verbose and ir is not None:
                if len(ir.signal.processOrder) > 0:
                    print("")
                    print("=== Signal Order ===")
                    for so in ir.signal.processOrder:
                        o = ir.objects[so.id]
                        if len(o.args) > 0:
                            print("{0} {{{1}}}".format(
                                o.type,
                                " ".join([f"{k}:{v}" for k, v in o.args.items()])))
                        else:
                            print(o.type)

        return CompilerResp(
            stage="hv2ir",
            compile_time=time.time() - tick,  # record the total compile time
            notifs=hv_graph.get_notices(),
            in_file=os.path.basename(hv_file),
            in_dir=os.path.dirname(hv_file),
            out_file=os.path.basename(ir_file),
            out_dir=os.path.dirname(ir_file),
            obj_counter=hv_counter,
            ir=ir
        )


def main() -> None:
    parser = argparse.ArgumentParser(
        description="A C-language compiler for the Heavy audio programming language.")
    parser.add_argument(
        "hv_path",
        help="The patch to the top-level patch to compile.")
    parser.add_argument(
        "--hv_ir_path",
        default="./heavy.hv.ir.json",
        help="The output path of the hv.ir.json file.")
    parser.add_argument(
        "--name",
        default="heavy",
        help="")
    parser.add_argument("-v", "--verbose", action="count")
    args = parser.parse_args()

    d = hv2ir.compile(
        hv_file=args.hv_path,
        ir_file=args.hv_ir_path,
        patch_name=args.name,
        verbose=args.verbose)

    if args.verbose:
        print(f"Total hv2ir time: {(d.compile_time * 1000):.2f}ms")


if __name__ == "__main__":
    main()



import jinja2
import os
import shutil
import time
import json2daisy  # type: ignore

from typing import Any, Dict, Optional

from ..copyright import copyright_manager
from . import parameters

from hvcc.interpreters.pd2hv.NotificationEnum import NotificationEnum
from hvcc.types.compiler import Generator, CompilerResp, CompilerNotif, CompilerMsg, ExternInfo
from hvcc.types.meta import Meta, Daisy


hv_midi_messages = {
    "__hv_noteout",
    "__hv_ctlout",
    "__hv_polytouchout",
    "__hv_pgmout",
    "__hv_touchout",
    "__hv_bendout",
    "__hv_midiout",
    "__hv_midioutport"
}


class c2daisy(Generator):
    """ Generates a Daisy wrapper for a given patch.
    """

    @classmethod
    def compile(
        cls,
        c_src_dir: str,
        out_dir: str,
        externs: ExternInfo,
        patch_name: Optional[str] = None,
        patch_meta: Meta = Meta(),
        num_input_channels: int = 0,
        num_output_channels: int = 0,
        copyright: Optional[str] = None,
        verbose: Optional[bool] = False
    ) -> CompilerResp:

        tick = time.time()

        out_dir = os.path.join(out_dir, "daisy")

        daisy_meta: Daisy = patch_meta.daisy
        board = daisy_meta.board

        copyright_c = copyright_manager.get_copyright_for_c(copyright)

        try:
            # ensure that the output directory does not exist
            out_dir = os.path.abspath(out_dir)
            if os.path.exists(out_dir):
                shutil.rmtree(out_dir)

            # copy over static files
            shutil.copytree(os.path.join(os.path.dirname(__file__), "static"), out_dir)

            # copy over generated C source files
            source_dir = os.path.join(out_dir, "source")
            shutil.copytree(c_src_dir, source_dir)

            if daisy_meta.board_file is not None:
                header, board_info = json2daisy.generate_header_from_file(daisy_meta.board_file)
            else:
                header, board_info = json2daisy.generate_header_from_name(board)

            # remove heavy out params from externs
            externs.parameters.outParam = [
                t for t in externs.parameters.outParam if not any(x == y for x in hv_midi_messages for y in t)]

            component_glue = parameters.parse_parameters(
                externs.parameters, board_info['components'], board_info['aliases'], 'hardware')
            component_glue['class_name'] = board_info['name']
            component_glue['patch_name'] = patch_name
            component_glue['header'] = f"HeavyDaisy_{patch_name}.hpp"
            component_glue['max_channels'] = board_info['channels']
            component_glue['num_output_channels'] = num_output_channels
            component_glue['has_midi'] = board_info['has_midi']
            component_glue['displayprocess'] = board_info['displayprocess']
            component_glue['debug_printing'] = daisy_meta.debug_printing
            component_glue['usb_midi'] = daisy_meta.usb_midi
            component_glue['pool_sizes_kb'] = externs.memoryPoolSizesKb

            # samplerate
            samplerate = daisy_meta.samplerate
            if samplerate >= 96000:
                component_glue['samplerate'] = 96000
            elif samplerate >= 48000:
                component_glue['samplerate'] = 48000
            elif samplerate >= 32000:
                component_glue['samplerate'] = 32000
            elif samplerate >= 16000:
                component_glue['samplerate'] = 16000
            else:
                component_glue['samplerate'] = 8000

            # blocksize
            blocksize = daisy_meta.blocksize
            if blocksize is not None:
                component_glue['blocksize'] = max(min(256, blocksize), 1)
            else:
                component_glue['blocksize'] = None

            component_glue['copyright'] = copyright_c

            daisy_h_path = os.path.join(source_dir, f"HeavyDaisy_{patch_name}.hpp")
            with open(daisy_h_path, "w") as f:
                f.write(header)

            loader = jinja2.FileSystemLoader(os.path.join(os.path.dirname(os.path.abspath(__file__)), 'templates'))
            env = jinja2.Environment(loader=loader, trim_blocks=True, lstrip_blocks=True)
            daisy_cpp_path = os.path.join(source_dir, f"HeavyDaisy_{patch_name}.cpp")

            rendered_cpp = env.get_template('HeavyDaisy.cpp').render(component_glue)
            with open(daisy_cpp_path, 'w') as f:
                f.write(rendered_cpp)

            makefile_replacements: Dict[str, Any] = {'name': patch_name}
            makefile_replacements['linker_script'] = daisy_meta.linker_script

            # libdaisy path
            path = daisy_meta.libdaisy_path
            if isinstance(path, int):
                makefile_replacements['libdaisy_path'] = f'{"../" * path}libdaisy'
            elif isinstance(path, str):
                makefile_replacements['libdaisy_path'] = path

            makefile_replacements['bootloader'] = daisy_meta.bootloader
            makefile_replacements['debug_printing'] = daisy_meta.debug_printing

            rendered_makefile = env.get_template('Makefile').render(makefile_replacements)
            with open(os.path.join(source_dir, "Makefile"), "w") as f:
                f.write(rendered_makefile)

            # ======================================================================================

            return CompilerResp(
                stage="c2daisy",
                in_dir=c_src_dir,
                out_dir=out_dir,
                out_file=os.path.basename(daisy_h_path),
                compile_time=time.time() - tick
            )

        except Exception as e:
            return CompilerResp(
                stage="c2daisy",
                notifs=CompilerNotif(
                    has_error=True,
                    exception=e,
                    warnings=[],
                    errors=[CompilerMsg(
                        enum=NotificationEnum.ERROR_EXCEPTION,
                        message=str(e)
                    )]
                ),
                in_dir=c_src_dir,
                out_dir=out_dir,
                compile_time=time.time() - tick
            )



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
# Copyright (C) 2023-2024 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Dict, Optional, List, Set, Tuple

from .HeavyIrObject import HeavyIrObject
from .HeavyGraph import HeavyGraph


class HIrTabwrite(HeavyIrObject):
    """ __tabwrite~f
    """

    def __init__(
        self,
        obj_type: str,
        args: Optional[Dict] = None,
        graph: Optional[HeavyGraph] = None,
        annotations: Optional[Dict] = None
    ) -> None:
        assert obj_type in {"__tabwrite~f", "__tabwrite_stoppable~f", "__tabwrite"}
        super().__init__(obj_type, args=args, graph=graph, annotations=annotations)

    def reduce(self) -> Optional[Tuple[Set, List]]:
        if self.graph is not None:
            table_obj = self.graph.resolve_object_for_name(
                self.args["table"],
                ["table", "__table"])
            if table_obj is not None:
                self.args["table_id"] = table_obj.id
                return ({self}, [])
            else:
                self.add_error(f"Can't find table with name \"{self.args['table']}\".")

        return None



# Heavy Compiler Collection
# Copyright (C) 2024 Wasted Audio
#
# SPDX-License-Identifier: GPL-3.0-only

from pydantic import BaseModel, RootModel
from typing import List, Optional, Dict, Literal, Union


LangConnectionType = Literal["-->", "-~>", "~f>"]
LangLetType = Literal["-->", "-~>", "~i>", "~f>", "signal"]
LangValueType = Literal["float", "int", "string", "bool", "dict", "floatarray", "intarray", "stringarray"]


class LangArg(BaseModel):
    name: str
    value_type: Optional[LangValueType] = None
    description: str
    default: Union[float, int, str, Dict, List, None] = None
    required: bool


class LangLet(BaseModel):
    name: str
    connectionType: LangConnectionType
    description: str


class LangNode(BaseModel):
    description: str
    inlets: List[LangLet]
    outlets: List[LangLet]
    args: List[LangArg]
    alias: List[str]
    tags: List[str]


class HeavyLangType(RootModel):
    root: Dict[str, LangNode]


if __name__ == "__main__":
    import json
    import importlib_resources

    heavy_lang_json = importlib_resources.files('hvcc') / 'core/json/heavy.lang.json'
    with open(heavy_lang_json, "r") as f:
        data = json.load(f)
        heavy_lang = HeavyLangType(root=data)
        print(heavy_lang.root.keys())



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
# Copyright (C) 2023 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Optional, List, Dict

from .PdObject import PdObject


class PdUnpackObject(PdObject):
    def __init__(
        self,
        obj_type: str,
        obj_args: Optional[List] = None,
        pos_x: int = 0,
        pos_y: int = 0
    ) -> None:
        assert obj_type == "unpack"
        super().__init__(obj_type, obj_args, pos_x, pos_y)

        if len(self.obj_args) == 0:
            self.add_error("Unpack requires at least one argument.")
        if not (set(self.obj_args) <= set(["f", "s"])):
            self.add_warning("Heavy only supports arguments 'f' and 's' to unpack.")

    def to_hv(self) -> Dict:
        """ Creates a graph dynamically based on the number of arguments.

            [inlet                                                ]
            |                        |      |
            [@hv_obj __slice 0 1]    ...    [@hv_obj __slice N-1 1]
            |                        |      |
            [outlet_0]               ...    [outlet_N-1]
        """

        hv_graph: Dict = {
            "type": "graph",
            "imports": [],
            "args": [],
            "objects": {
                "inlet": {
                    "type": "inlet",
                    "args": {
                        "type": "-->",
                        "index": 0
                    },
                    "properties": {"x": 0, "y": 0}
                }
            },
            "connections": [],
            "properties": {"x": self.pos_x, "y": self.pos_y}
        }

        # NOTE(mhroth): reverse the iteration such that connections are
        # added in the correct order
        for i in reversed(range(len(self.obj_args))):
            # add slices to graph
            hv_graph["objects"][f"slice_{i}"] = {
                "type": "slice",
                "args": {
                    "index": i,
                    "length": 1
                },
                "properties": {"x": 0, "y": 0}
            }

            # add outlets to graph
            hv_graph["objects"][f"outlet_{i}"] = {
                "type": "outlet",
                "args": {
                    "type": "-->",
                    "index": i
                },
                "properties": {"x": 0, "y": 0}
            }

            # add connection from inlet to slice
            hv_graph["connections"].append({
                "from": {"id": "inlet", "outlet": 0},
                "to": {"id": f"slice_{i}", "inlet": 0},
                "type": "-->"
            })

            # add connection from slice to outlet
            hv_graph["connections"].append({
                "from": {"id": f"slice_{i}", "outlet": 0},
                "to": {"id": f"outlet_{i}", "inlet": 0},
                "type": "-->"
            })

        return hv_graph



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
# Copyright (C) 2023-2024 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Dict, List

from .HeavyObject import HeavyObject

from hvcc.types.IR import IRSignalList


class SignalLorenz(HeavyObject):

    c_struct = "SignalLorenz"
    preamble = "sLorenz"

    @classmethod
    def get_C_header_set(cls) -> set:
        return {"HvSignalLorenz.h"}

    @classmethod
    def get_C_file_set(cls) -> set:
        return {"HvSignalLorenz.h", "HvSignalLorenz.c", "HvMath.h"}

    @classmethod
    def get_C_init(cls, obj_type: str, obj_id: str, args: Dict) -> List[str]:
        return [
            "sLorenz_init(&sLorenz_{0}, {1}f, {2}f, {3}f);".format(
                obj_id,
                float(args["x"]),
                float(args["y"]),
                float(args["z"]))
        ]

    @classmethod
    def get_C_free(cls, obj_type: str, obj_id: str, args: Dict) -> List[str]:
        return []  # nothing to free

    @classmethod
    def get_C_onMessage(cls, obj_type: str, obj_id: str, inlet_index: int, args: Dict) -> List[str]:
        return [
            "sLorenz_onMessage(_c, &Context(_c)->sLorenz_{0}, {1}, m);".format(
                obj_id,
                inlet_index)
        ]

    @classmethod
    def get_C_process(cls, process_dict: IRSignalList, obj_type: str, obj_id: str, args: Dict) -> List[str]:
        return [
            "__hv_lorenz_f(&sLorenz_{0}, {1}, {2});".format(
                process_dict.id,
                ", ".join(["VIf({0})".format(cls._c_buffer(b)) for b in process_dict.inputBuffers]),
                ", ".join(["VOf({0})".format(cls._c_buffer(b)) for b in process_dict.outputBuffers])
            )
        ]



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
# Copyright (C) 2023-2024 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import decimal
import json
import importlib_resources
from typing import Optional, List, Dict, Any, Union, cast

from .Connection import Connection
from .NotificationEnum import NotificationEnum
from .PdObject import PdObject

from hvcc.types.IR import HeavyIRType, IRNode, IRArg
from hvcc.types.Lang import HeavyLangType, LangNode, LangArg


class HeavyObject(PdObject):

    heavy_lang_json = importlib_resources.files('hvcc') / 'core/json/heavy.lang.json'
    with open(heavy_lang_json, "r") as f:
        __HEAVY_LANG_OBJS = HeavyLangType(**json.load(f)).root

    heavy_ir_json = importlib_resources.files('hvcc') / 'core/json/heavy.ir.json'
    with open(heavy_ir_json, "r") as f:
        __HEAVY_IR_OBJS = HeavyIRType(**json.load(f)).root

    def __init__(
        self,
        obj_type: str,
        obj_args: Optional[List] = None,
        pos_x: int = 0,
        pos_y: int = 0
    ) -> None:
        super().__init__(obj_type, obj_args, pos_x, pos_y)

        self.__obj_dict: Union[IRNode, LangNode]

        # get the object dictionary (note that it is NOT a copy)
        if self.is_hvlang:
            self.__obj_dict = self.__HEAVY_LANG_OBJS[obj_type]
        elif self.is_hvir:
            self.__obj_dict = self.__HEAVY_IR_OBJS[obj_type]
        else:
            assert False, f"{obj_type} is not a Heavy Lang or IR object."

        # resolve arguments
        obj_args = obj_args or []
        self.obj_dict = {}

        for i, a in enumerate(self.__obj_dict.args):
            arg = cast(Union[IRArg, LangArg], a)
            # if the argument exists (and has been correctly resolved)
            if i < len(obj_args) and obj_args[i] is not None:
                # force the Heavy argument type
                # Catch type errors as early as possible
                try:
                    self.obj_dict[arg.name] = self.force_arg_type(
                        obj_args[i],
                        arg.value_type)
                except Exception as e:
                    self.add_error(
                        f"Heavy {obj_type} cannot convert argument \"{arg.name}\""
                        f" with value \"{obj_args[i]}\" to type {arg.value_type}: {e}")
            else:
                # the default argument is required
                if arg.required:
                    self.add_error(
                        f"Required argument \"{arg.name}\" to object {obj_type} not present: {obj_args}")
                else:
                    # don't worry about supplying a default,
                    # let hv2ir take care of it. pd2hv only passes on the
                    # provided parameters.
                    # self.obj_args[a["name"]] = a["default"]
                    pass

        self.__annotations = {}

        # send/receive, table, etc. must have public scope
        # TODO(mhroth): dirty
        if obj_type in ["table", "__table", "send", "__send", "receive", "__receive"]:
            self.__annotations["scope"] = "public"

    @classmethod
    def force_arg_type(
        cls,
        value: str,
        value_type: Optional[str] = None
    ) -> Any:
        # TODO(mhroth): add support for mixedarray?
        if value_type == "auto":
            try:
                return float(value)
            except Exception:
                return str(value)
        elif value_type == "float":
            return float(value)
        elif value_type == "int":
            return int(decimal.Decimal(value))
        elif value_type == "string":
            return str(value)
        elif value_type == "boolean":
            if isinstance(value, str):
                return value.strip().lower() not in {"false", "f", "0"}
            else:
                return bool(value)
        elif value_type == "floatarray":
            if isinstance(value, list):
                return [float(v) for v in value]
            if isinstance(value, str):
                return [float(v) for v in value.split()]
            else:
                raise Exception(f"Cannot convert value to type floatarray: {value}")
        elif value_type == "intarray":
            if isinstance(value, list):
                return [int(v) for v in value]
            if isinstance(value, str):
                return [int(v) for v in value.split()]
            else:
                raise Exception(f"Cannot convert value to type intarray: {value}")
        elif value_type == "stringarray":
            if isinstance(value, list):
                return [str(v) for v in value]
            if isinstance(value, str):
                return [str(v) for v in value.split()]
            else:
                raise Exception(f"Cannot convert value to type stringarray: {value}")
        else:
            # NOTE(mhroth): if value_type is not a known type or None, that is
            # not necessarily an error. It may simply be that the value should
            # not be resolved to anything other than what it already is.
            # This happens most often with message objects.
            return value

    @property
    def is_hvlang(self) -> bool:
        return self.obj_type in self.__HEAVY_LANG_OBJS.keys()

    @property
    def is_hvir(self) -> bool:
        return self.obj_type in self.__HEAVY_IR_OBJS.keys()

    def get_inlet_connection_type(self, inlet_index: int) -> Optional[str]:
        """ Returns the inlet connection type, None if the inlet does not exist.
        """
        # TODO(mhroth): it's stupid that hvlang and hvir json have different data formats here
        if self.is_hvlang and isinstance(self.__obj_dict, LangNode):
            if len(self.__obj_dict.inlets) > inlet_index:
                return self.__obj_dict.inlets[inlet_index].connectionType
            else:
                return None
        elif self.is_hvir and isinstance(self.__obj_dict, IRNode):
            if len(self.__obj_dict.inlets) > inlet_index:
                return self.__obj_dict.inlets[inlet_index]
            else:
                return None
        else:
            return None

    def get_outlet_connection_type(self, outlet_index: int) -> Optional[str]:
        """ Returns the outlet connection type, None if the inlet does not exist.
        """
        # TODO(mhroth): it's stupid that hvlang and hvir json have different data formats here
        if self.is_hvlang and isinstance(self.__obj_dict, LangNode):
            if len(self.__obj_dict.outlets) > outlet_index:
                return self.__obj_dict.outlets[outlet_index].connectionType
            else:
                return None
        elif self.is_hvir and isinstance(self.__obj_dict, IRNode):
            if len(self.__obj_dict.outlets) > outlet_index:
                return self.__obj_dict.outlets[outlet_index]
            else:
                return None
        else:
            return None

    def add_connection(self, c: Connection) -> None:
        """ Adds a connection, either inlet or outlet, to this object.
        """
        if c.from_id == self.obj_id:
            if self.get_outlet_connection_type(c.outlet_index) is not None:
                self._outlet_connections[str(c.outlet_index)].append(c)
            else:
                self.add_error(
                    f"Connection made from non-existent outlet at {self.obj_type}:{c.outlet_index}.",
                    enum=NotificationEnum.ERROR_UNABLE_TO_CONNECT_OBJECTS)
        elif c.to_id == self.obj_id:
            if self.get_inlet_connection_type(c.inlet_index) is not None:
                self._inlet_connections[str(c.inlet_index)].append(c)
            else:
                self.add_error(
                    f"Connection made to non-existent inlet at [{self.obj_type} {self.obj_dict}]:{c.inlet_index}.",
                    enum=NotificationEnum.ERROR_UNABLE_TO_CONNECT_OBJECTS)
        else:
            raise Exception("Adding a connection to the wrong object!")

    def to_hv(self) -> Dict:
        return {
            "type": self.obj_type,
            "args": self.obj_dict,
            "properties": {
                "x": self.pos_x,
                "y": self.pos_y
            },
            "annotations": self.__annotations
        }



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
# Copyright (C) 2023-2024 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Dict, List

from .HeavyObject import HeavyObject


class ControlTabread(HeavyObject):

    c_struct = "ControlTabread"
    preamble = "cTabread"

    @classmethod
    def get_C_header_set(cls) -> set:
        return {"HvControlTabread.h"}

    @classmethod
    def get_C_file_set(cls) -> set:
        return {"HvControlTabread.h", "HvControlTabread.c"}

    @classmethod
    def get_C_init(cls, obj_type: str, obj_id: str, args: Dict) -> List[str]:
        return [
            "cTabread_init(&cTabread_{0}, &hTable_{1}); // {2}".format(
                obj_id,
                args["table_id"],
                args["table"])
        ]

    @classmethod
    def get_C_free(cls, obj_type: str, obj_id: str, args: Dict) -> List[str]:
        return []

    @classmethod
    def get_C_onMessage(cls, obj_type: str, obj_id: str, inlet_index: int, args: Dict) -> List[str]:
        return [
            "cTabread_onMessage(_c, &Context(_c)->cTabread_{0}, {1}, m, &cTabread_{0}_sendMessage);".format(
                obj_id,
                inlet_index)
        ]



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
# Copyright (C) 2023-2024 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import jinja2
import json
import os
import shutil
import time

from collections import Counter
from collections import OrderedDict
from typing import Dict, List, Optional, Type, Union

from hvcc.generators.ir2c.PrettyfyC import PrettyfyC
from hvcc.generators.copyright import copyright_manager

from hvcc.generators.ir2c.ControlBinop import ControlBinop
from hvcc.generators.ir2c.ControlCast import ControlCast
from hvcc.generators.ir2c.ControlDelay import ControlDelay
from hvcc.generators.ir2c.ControlIf import ControlIf
from hvcc.generators.ir2c.ControlMessage import ControlMessage
from hvcc.generators.ir2c.ControlPack import ControlPack
from hvcc.generators.ir2c.ControlPrint import ControlPrint
from hvcc.generators.ir2c.ControlReceive import ControlReceive
from hvcc.generators.ir2c.ControlRandom import ControlRandom
from hvcc.generators.ir2c.ControlSend import ControlSend
from hvcc.generators.ir2c.ControlSlice import ControlSlice
from hvcc.generators.ir2c.ControlSwitchcase import ControlSwitchcase
from hvcc.generators.ir2c.ControlSystem import ControlSystem
from hvcc.generators.ir2c.ControlTabhead import ControlTabhead
from hvcc.generators.ir2c.ControlTabread import ControlTabread
from hvcc.generators.ir2c.ControlTabwrite import ControlTabwrite
from hvcc.generators.ir2c.ControlUnop import ControlUnop
from hvcc.generators.ir2c.ControlVar import ControlVar
from hvcc.generators.ir2c.HeavyObject import HeavyObject
from hvcc.generators.ir2c.HeavyTable import HeavyTable
from hvcc.generators.ir2c.SignalConvolution import SignalConvolution
from hvcc.generators.ir2c.SignalBiquad import SignalBiquad
from hvcc.generators.ir2c.SignalCPole import SignalCPole
from hvcc.generators.ir2c.SignalDel1 import SignalDel1
from hvcc.generators.ir2c.SignalEnvelope import SignalEnvelope
from hvcc.generators.ir2c.SignalLine import SignalLine
from hvcc.generators.ir2c.SignalLorenz import SignalLorenz
from hvcc.generators.ir2c.SignalMath import SignalMath
from hvcc.generators.ir2c.SignalPhasor import SignalPhasor
from hvcc.generators.ir2c.SignalRPole import SignalRPole
from hvcc.generators.ir2c.SignalSample import SignalSample
from hvcc.generators.ir2c.SignalSamphold import SignalSamphold
from hvcc.generators.ir2c.SignalTabhead import SignalTabhead
from hvcc.generators.ir2c.SignalTabread import SignalTabread
from hvcc.generators.ir2c.SignalTabwrite import SignalTabwrite
from hvcc.generators.ir2c.SignalVar import SignalVar

from hvcc.types.compiler import CompilerResp, ExternInfo
from hvcc.types.IR import IRGraph


class ir2c:

    __OBJECT_CLASS_DICT = {
        "__delay": ControlDelay,
        "__if": ControlIf,
        "__print": ControlPrint,
        "__random": ControlRandom,
        "__var": ControlVar,
        "__table": HeavyTable,
        "__cast_b": ControlCast,
        "__cast_f": ControlCast,
        "__cast_s": ControlCast,
        "__message": ControlMessage,
        "__system": ControlSystem,
        "__receive": ControlReceive,
        "__switchcase": ControlSwitchcase,
        "__conv~f": SignalConvolution,
        "__biquad~f": SignalBiquad,
        "__biquad_k~f": SignalBiquad,
        "__env~f": SignalEnvelope,
        "__line~f": SignalLine,
        "__lorenz~f": SignalLorenz,
        "__del1~f": SignalDel1,
        "__tabread~if": SignalTabread,
        "__tabread~f": SignalTabread,
        "__tabreadu~f": SignalTabread,
        "__tabread_stoppable~f": SignalTabread,
        "__tabhead~f": SignalTabhead,
        "__tabwrite~f": SignalTabwrite,
        "__tabwrite_stoppable~f": SignalTabwrite,
        "__phasor~f": SignalPhasor,
        "__phasor_k~f": SignalPhasor,
        "__sample~f": SignalSample,
        "__samphold~f": SignalSamphold,
        "__slice": ControlSlice,
        "__send": ControlSend,
        "__tabhead": ControlTabhead,
        "__tabread": ControlTabread,
        "__tabwrite": ControlTabwrite,
        "__pack": ControlPack,
        "__rpole~f": SignalRPole,
        "__cpole~f": SignalCPole
    }

    # the base set of C files necessary for the patch
    __BASE_FILE_SET = {
        "HeavyContextInterface.hpp",
        "HeavyContext.hpp", "HeavyContext.cpp",
        "HvHeavy.h", "HvHeavyInternal.h", "HvHeavy.cpp",
        "HvUtils.h", "HvUtils.c", "HvMath.h",
        "HvMessageQueue.h", "HvMessageQueue.c",
        "HvMessagePool.h", "HvMessagePool.c",
        "HvTable.h", "HvTable.c",
        "HvMessage.h", "HvMessage.c",
        "HvLightPipe.h", "HvLightPipe.c"
    }

    @classmethod
    def filter_hvhash(cls, x: Union[float, str]) -> str:
        """ Return the hash string of an object.
        """
        return HeavyObject.get_hash_string(x)

    @classmethod
    def filter_extern(cls, d: Dict) -> Dict:
        """ Return a dictionary of objects that are externed.
        """
        return {k: v for k, v in d.items() if v.extern}

    @classmethod
    def get_class(cls, obj_type: str) -> Type[HeavyObject]:
        if SignalMath.handles_type(obj_type):
            return SignalMath
        elif ControlBinop.handles_type(obj_type):
            return ControlBinop
        elif ControlUnop.handles_type(obj_type):
            return ControlUnop
        elif SignalVar.handles_type(obj_type):
            return SignalVar
        elif obj_type in ir2c.__OBJECT_CLASS_DICT:
            return ir2c.__OBJECT_CLASS_DICT[obj_type]
        else:
            raise Exception(f"No class found for object type \"{obj_type}\".")

    @classmethod
    def compile(
        cls,
        hv_ir_path: str,
        static_dir: str,
        output_dir: str,
        externs: ExternInfo,
        copyright: Optional[str] = None,
        nodsp: Optional[bool] = False
    ) -> CompilerResp:
        """ Compiles a HeavyIR file into a C.
            Returns a tuple of compile time in seconds, a notification dictionary,
            and a HeavyIR object counter.
        """

        # keep track of the total compile time
        tick = time.time()

        # establish the jinja environment
        env = jinja2.Environment()
        env.filters["hvhash"] = cls.filter_hvhash
        env.filters["extern"] = cls.filter_extern
        env.loader = jinja2.FileSystemLoader(
            os.path.join(os.path.dirname(__file__), "templates"))

        # read the hv.ir.json file
        with open(hv_ir_path, "r") as f:
            ir = IRGraph(**json.load(f))

        # generate the copyright
        copyright = copyright_manager.get_copyright_for_c(copyright)

        #
        # Parse the hv.ir data structure and generate C-language strings.
        #

        # generate set of header files to include
        include_set = set([x for o in ir.objects.values() for x in ir2c.get_class(o.type).get_C_header_set()])

        # generate set of files to add to project
        file_set = set([x for o in ir.objects.values() for x in ir2c.get_class(o.type).get_C_file_set()])
        file_set.update(ir2c.__BASE_FILE_SET)

        # generate object definition and initialisation list
        init_list: List = []
        free_list = []
        def_list = []
        decl_list = []
        for obj_id in ir.init.order:
            o = ir.objects[obj_id]
            obj_class = ir2c.get_class(o.type)
            init_list.extend(obj_class.get_C_init(o.type, obj_id, o.args))
            def_list.extend(obj_class.get_C_def(o.type, obj_id))
            free_list.extend(obj_class.get_C_free(o.type, obj_id, o.args))

        impl_list = []
        for msg in ir.control.sendMessage:
            obj_id = msg.id
            o = ir.objects[obj_id]
            obj_class = ir2c.get_class(o.type)
            impl = obj_class.get_C_impl(
                o.type,
                obj_id,
                msg.onMessage,
                ir2c.get_class,
                ir.objects)
            impl_list.append("\n".join(PrettyfyC.prettyfy_list(impl)))
            decl_list.extend(obj_class.get_C_decl(o.type, obj_id, o.args))

        # generate static table data initialisers
        table_data_list: List = []
        for k, v in ir.tables.items():
            o = ir.objects[v.id]
            obj_class = ir2c.get_class(o.type)
            table_data_list.extend(obj_class.get_table_data_decl(
                o.type,
                v.id,
                o.args))

        # generate the list of functions to process
        process_list: List = []
        for sig in ir.signal.processOrder:
            obj_id = sig.id
            o = ir.objects[obj_id]
            process_list.extend(ir2c.get_class(o.type).get_C_process(
                sig,
                o.type,
                obj_id,
                o.args))

        #
        # Load the C-language template files and use the parsed strings to fill them in.
        #

        # make the output directory if necessary
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)

        # the project name to be used as a part of file and function names
        name = ir.name.escaped

        # ensure that send_receive dictionary is alphabetised by the receiver key
        send_receive = OrderedDict(sorted([(k, v) for k, v in ir.control.receivers.items()], key=lambda x: x[0]))

        # write HeavyContext.h
        with open(os.path.join(output_dir, f"Heavy_{name}.hpp"), "w") as f:
            f.write(env.get_template("Heavy_NAME.hpp").render(
                name=name,
                include_set=include_set,
                decl_list=decl_list,
                def_list=def_list,
                signal=ir.signal,
                copyright=copyright,
                externs=externs))

        # write C++ implementation
        with open(os.path.join(output_dir, f"Heavy_{name}.cpp"), "w") as f:
            f.write(env.get_template("Heavy_NAME.cpp").render(
                name=name,
                signal=ir.signal,
                init_list=init_list,
                free_list=free_list,
                impl_list=impl_list,
                send_receive=send_receive,
                send_table=ir.tables,
                process_list=process_list,
                table_data_list=table_data_list,
                copyright=copyright,
                nodsp=nodsp))

        # write C API, hv_NAME.h
        with open(os.path.join(output_dir, f"Heavy_{name}.h"), "w") as f:
            f.write(env.get_template("Heavy_NAME.h").render(
                name=name,
                copyright=copyright,
                externs=externs))

        # copy static files to output directory
        for f in file_set:
            shutil.copy2(
                src=os.path.join(static_dir, str(f)),
                dst=os.path.join(output_dir, str(f)))

        # generate HeavyIR object counter
        ir_counter = Counter([obj.type for obj in ir.objects.values()])

        return CompilerResp(
            stage="ir2c",
            in_dir=os.path.dirname(hv_ir_path),
            in_file=os.path.basename(hv_ir_path),
            out_dir=output_dir,
            compile_time=(time.time() - tick),
            obj_counter=ir_counter
        )


def main() -> None:
    parser = argparse.ArgumentParser(
        description="A Heavy.IR to C-language translator.")
    parser.add_argument(
        "hv_ir_path",
        help="The path to the Heavy.IR file to read.")
    parser.add_argument(
        "--static_dir",
        default="./static",
        help="The path to the static C files.")
    parser.add_argument(
        "--output_dir",
        default="./out",
        help="")
    parser.add_argument(
        "--copyright",
        default=None,
        help="A string indicating the owner of the copyright.")
    parser.add_argument("-v", "--verbose", action="count")
    args = parser.parse_args()

    externs = ExternInfo()

    results = ir2c.compile(
        args.hv_ir_path,
        args.static_dir,
        args.output_dir,
        externs,
        args.copyright)

    if args.verbose:
        print("Total ir2c time: {0:.2f}ms".format(results.compile_time * 1000))


if __name__ == "__main__":
    main()



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
# Copyright (C) 2023-2024 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Dict, List
from .HeavyObject import HeavyObject


class ControlBinop(HeavyObject):
    # a dictionary translating from the operation argument to the C case
    __OPERATION_DICT = {
        "__add": "HV_BINOP_ADD",
        "__add_k": "HV_BINOP_ADD",
        "__sub": "HV_BINOP_SUBTRACT",
        "__sub_k": "HV_BINOP_SUBTRACT",
        "__mul": "HV_BINOP_MULTIPLY",
        "__mul_k": "HV_BINOP_MULTIPLY",
        "__div": "HV_BINOP_DIVIDE",
        "__div_k": "HV_BINOP_DIVIDE",
        "__intdiv": "HV_BINOP_INT_DIV",
        "__intdiv_k": "HV_BINOP_INT_DIV",
        "__shiftleft": "HV_BINOP_BIT_LEFTSHIFT",
        "__shiftleft_k": "HV_BINOP_BIT_LEFTSHIFT",
        "__shiftright": "HV_BINOP_BIT_RIGHTSHIFT",
        "__shiftright_k": "HV_BINOP_BIT_RIGHTSHIFT",
        "__and": "HV_BINOP_BIT_AND",
        "__and_k": "HV_BINOP_BIT_AND",
        "__or": "HV_BINOP_BIT_OR",
        "__or_k": "HV_BINOP_BIT_OR",
        "__^": "HV_BINOP_BIT_XOR",
        "__==": "HV_BINOP_EQL_EQL",
        "__!=": "HV_BINOP_NOT_EQL",
        "__logand": "HV_BINOP_LOGICAL_AND",
        "__logand_k": "HV_BINOP_LOGICAL_AND",
        "__logor": "HV_BINOP_LOGICAL_OR",
        "__logor_k": "HV_BINOP_LOGICAL_OR",
        "__eq": "HV_BINOP_EQ",
        "__eq_k": "HV_BINOP_EQ",
        "__neq": "HV_BINOP_NEQ",
        "__neq_k": "HV_BINOP_NEQ",
        "__lt": "HV_BINOP_LESS_THAN",
        "__lt_k": "HV_BINOP_LESS_THAN",
        "__lte": "HV_BINOP_LESS_THAN_EQL",
        "__lte_k": "HV_BINOP_LESS_THAN_EQL",
        "__gt": "HV_BINOP_GREATER_THAN",
        "__gt_k": "HV_BINOP_GREATER_THAN",
        "__gte": "HV_BINOP_GREATER_THAN_EQL",
        "__gte_k": "HV_BINOP_GREATER_THAN_EQL",
        "__max": "HV_BINOP_MAX",
        "__max_k": "HV_BINOP_MAX",
        "__min": "HV_BINOP_MIN",
        "__min_k": "HV_BINOP_MIN",
        "__pow": "HV_BINOP_POW",
        "__pow_k": "HV_BINOP_POW",
        "__atan2": "HV_BINOP_ATAN2",
        "__atan2_k": "HV_BINOP_ATAN2",
        "__unimod": "HV_BINOP_MOD_UNIPOLAR",
        "__unimod_k": "HV_BINOP_MOD_UNIPOLAR",
        "__bimod": "HV_BINOP_MOD_BIPOLAR",
        "__bimod_k": "HV_BINOP_MOD_BIPOLAR"
    }

    c_struct = "ControlBinop"
    preamble = "cBinop"

    @classmethod
    def handles_type(cls, obj_type: str) -> bool:
        """Returns true if the object type can be handled by this class
        """
        return obj_type in cls.__OPERATION_DICT

    @classmethod
    def get_C_header_set(cls) -> set:
        return {"HvControlBinop.h"}

    @classmethod
    def get_C_file_set(cls) -> set:
        return {"HvControlBinop.h", "HvControlBinop.c"}

    @classmethod
    def get_C_init(cls, obj_type: str, obj_id: str, args: Dict) -> List[str]:
        if obj_type.endswith("_k"):
            return []
        else:
            obj_arg = float(list(args.values())[0])
            return [f"cBinop_init(&cBinop_{obj_id}, {obj_arg}f); // {obj_type}"]

    @classmethod
    def get_C_free(cls, obj_type: str, obj_id: str, args: Dict) -> List[str]:
        return []  # no need to free any control binop objects

    @classmethod
    def get_C_onMessage(cls, obj_type: str, obj_id: str, inlet_index: int, args: Dict) -> List[str]:
        if obj_type.endswith("_k"):
            return [
                "cBinop_k_onMessage(_c, NULL, {0}, {1}f, {2}, m, &cBinop_{3}_sendMessage);".format(
                    cls.__OPERATION_DICT[obj_type[:-2]],
                    float(args["k"]),
                    inlet_index,
                    obj_id)
            ]
        else:
            return [
                "cBinop_onMessage(_c, &Context(_c)->cBinop_{0}, {1}, {2}, m, &cBinop_{0}_sendMessage);".format(
                    obj_id,
                    cls.__OPERATION_DICT[obj_type],
                    inlet_index)
            ]




from copy import deepcopy
from typing import Any, Dict, Optional

from hvcc.types.compiler import ExternParams


def filter_match(
    set: Dict,
    key: str,
    match: str,
    key_exclude: Optional[str] = None,
    match_exclude: Optional[str] = None
) -> filter:
    if (key_exclude is not None and match_exclude is not None):
        return filter(lambda x: x.get(key, '') == match and x.get(key_exclude, '') != match_exclude, set)
    else:
        return filter(lambda x: x.get(key, '') == match, set)


def verify_param_exists(name: str, original_name: str, components: Dict, input: bool = True) -> Any:
    for comp in components:

        # Dealing with the cvouts the way we have it set up is really annoying
        if comp['component'] == 'CVOuts':

            if name == comp['name']:
                if input:
                    raise TypeError(
                        f'Parameter "{original_name}" cannot be used as an {"input" if input else "output"}')
                return
        else:
            variants = [mapping['name'].format_map(
                {'name': comp['name']}) for mapping in comp['mapping']]
            if name in variants:
                if input and comp['direction'] == 'output' or not input and comp['direction'] == 'input':
                    raise TypeError(
                        f'Parameter "{original_name}" cannot be used as an {"input" if input else "output"}')
                return

    raise NameError(f'Unknown parameter "{original_name}"')


def verify_param_direction(name: str, components: Dict) -> bool:
    for comp in components:
        if comp['component'] == 'CVOuts':
            if name == comp['name']:
                return True
        else:
            variants = [mapping['name'].format_map(
                {'name': comp['name']}) for mapping in comp['mapping']]
            if name in variants:
                return True
    return False


def get_root_component(variant: str, original_name: str, components: Dict) -> str:
    for comp in components:
        if comp['component'] == 'CVOuts':
            if variant == comp['name']:
                return variant
        else:
            variants = [mapping['name'].format_map(
                {'name': comp['name']}) for mapping in comp['mapping']]
            if variant in variants:
                return comp['name']
    raise NameError(f'Unknown parameter "{original_name}"')


def get_component_mapping(
    component_variant: str,
    original_name: str,
    component: Dict,
    components: Dict
) -> Dict:
    for variant in component['mapping']:
        if component['component'] == 'CVOuts':
            stripped = variant['name'].format_map({'name': ''})
            if stripped in component['name']:
                return variant
        elif variant['name'].format_map({'name': component['name']}) == component_variant:
            return variant
    raise NameError(f'Unknown parameter "{original_name}"')


def verify_param_used(
    component: Dict,
    params_in: Dict,
    params_out: Dict,
    params_in_original_name: Dict,
    params_out_original_name: Dict,
    components: Dict
) -> bool:
    # Exclude parents, since they don't have 1-1 i/o mapping
    if component.get('is_parent', False):
        return True
    for param in {**params_in, **params_out}:
        root = get_root_component(
            param, ({**params_in_original_name, **params_out_original_name})[param], components)
        if root == component['name']:
            return True
    return False


def de_alias(
    name: str,
    aliases: Dict,
    components: Dict
) -> str:
    low = name.lower()
    # simple case
    if low in aliases:
        return aliases[low]
    # aliased variant
    potential_aliases = list(filter(lambda x: x in low, aliases))
    for alias in potential_aliases:
        try:
            target_component = list(filter_match(
                components, 'name', aliases[alias]))[0]
            # The CVOuts setup really bothers me
            if target_component['component'] != 'CVOuts':
                for mapping in target_component['mapping']:
                    if mapping['name'].format_map({'name': alias}) == low:
                        return mapping['name'].format_map({'name': aliases[alias]})
        except IndexError:
            # No matching alias from filter
            pass
    # otherwise, it's a direct parameter or unkown one
    return low


def parse_parameters(
    parameters: ExternParams,
    components: Dict,
    aliases: Dict,
    object_name: str
) -> Dict:
    """
    Parses the `parameters` passed from hvcc and generates getters and setters
    according to the info in `components`. The `aliases` help disambiguate parameters
    and the `object_name` sets the identifier for the generated Daisy hardware class,
    in this case `hardware`.
    """

    # Verify that the params are valid and remove unused components
    replacements: Dict = {}
    params_in: Dict = {}
    params_in_original_names: Dict = {}
    for key, recv in parameters.inParam:
        de_aliased = de_alias(key, aliases, components)
        params_in[de_aliased] = recv
        params_in_original_names[de_aliased] = key

    params_out = {}
    params_out_original_names = {}
    for key, msg in parameters.outParam:
        de_aliased = de_alias(key, aliases, components)
        params_out[de_aliased] = msg
        params_out_original_names[de_aliased] = key

    [verify_param_exists(key, params_in_original_names[key],
                         components, input=True) for key in params_in]
    [verify_param_exists(key, params_out_original_names[key],
                         components, input=False) for key in params_out]

    for i in range(len(components) - 1, -1, -1):
        if not verify_param_used(
                components[i], params_in, params_out,
                params_in_original_names, params_out_original_names,
                components):
            components.pop(i)

    out_idx = 0

    replacements['parameters'] = []
    replacements['output_parameters'] = []
    replacements['loop_write_in'] = []
    replacements['callback_write_out'] = []
    replacements['loop_write_out'] = []
    replacements['hook_write_out'] = []
    replacements['callback_write_in'] = []

    for param_name, param in params_in.items():
        root = get_root_component(
            param_name, params_in_original_names[param_name], components)
        component = list(filter_match(components, 'name', root))[0]
        param_struct = {
            "hash_enum": params_in_original_names[param_name], 'name': root, 'type': component['component'].upper()}
        replacements['parameters'].append(param_struct)
        mapping = get_component_mapping(
            param_name, params_in_original_names[param_name], component, components)

        write_location = 'callback_write_in' if mapping.get(
            'where', 'callback') == 'callback' else 'loop_write_in'
        component_info = deepcopy(component)
        component_info['name'] = root
        component_info['class_name'] = object_name
        # A bit of a hack to get cv_1, etc to be written as CV_1
        component_info['name_upper'] = root.upper()
        component_info['value'] = f'output_data[{out_idx}]'
        component_info['default_prefix'] = component.get(
            "default_prefix", '') if component.get('default', False) else ''
        process = mapping["get"].format_map(component_info)

        replacements[write_location].append(
            {"process": process, "bool": mapping.get('bool', False),
                "hash_enum": params_in_original_names[param_name]})

    for param_name, param in params_out.items():
        root = get_root_component(
            param_name, params_out_original_names[param_name], components)
        component = list(filter_match(components, 'name', root))[0]

        mapping = get_component_mapping(
            param_name, params_out_original_names[param_name], component, components)

        default_prefix = component.get(
            "default_prefix", '') if component.get('default', False) else ''

        write_locations = {'callback': 'callback_write_out',
                           'loop': 'loop_write_out',
                           'hook': 'hook_write_out'}

        write_location = write_locations.get(mapping.get('where', 'callback'), 'callback_write_out')

        param_struct = {
            'hash_enum': params_out_original_names[param_name],
            'index': out_idx, 'name': param_name,
            'hook': write_location == 'hook_write_out'}
        replacements['output_parameters'].append(param_struct)

        component_info = deepcopy(component)
        component_info['hash_enum'] = params_out_original_names[param_name]
        component_info['name'] = root
        component_info['class_name'] = object_name
        component_info['value'] = f'output_data[{out_idx}]' if \
            write_location != 'hook_write_out' else 'sig'
        component_info['default_prefix'] = default_prefix
        write = mapping["set"].format_map(component_info)

        replacements[write_location].append(
            {"name": param_name,
             "process": write,
             "bool": mapping.get('bool', False),
             "value": component_info['value']})

        out_idx += 1

    replacements['output_comps'] = len(replacements['output_parameters'])

    return replacements



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
# Copyright (C) 2023 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from typing import List

from .Connection import Connection
from .HeavyObject import HeavyObject
from .PdGraph import PdGraph


class PdLibSignalGraph(PdGraph):

    def __init__(
        self,
        obj_args: List,
        pd_path: str,
        pos_x: int = 0,
        pos_y: int = 0
    ) -> None:
        super().__init__(obj_args, pd_path, pos_x, pos_y)

    def validate_configuration(self) -> None:
        """ Auto-detect control only connections to expected signal inlets and
            insert a sig~ (var) object inbetween to ensure it doesn't break.
        """

        # Note(joe): only checking first inlet, should we check all inlets by
        # default? i.e. samphold~ has a 2nd inlet that requires checking
        conns = self._inlet_connections.get("0", [])

        # only do this if no signal connections are present
        if (len([c for c in conns if c.conn_type == "~f>"]) == 0) and len(conns) > 0:

            # add sig~ object to parent graph
            if self.parent_graph is not None:
                sig_obj = HeavyObject(
                    obj_type="var",
                    obj_args=[0],
                    pos_x=int(self.pos_x),
                    pos_y=int(self.pos_y - 5))  # shift upwards a few points
                self.parent_graph.add_object(sig_obj)

                # add connection from sig~ to this abstraction
                c = Connection(sig_obj, 0, self, 0, "~f>")
                self.parent_graph._PdGraph__connections.append(c)  # update the local connections list
                sig_obj.add_connection(c)
                self.add_connection(c)

                # retrieve all control connections
                control_conns = [c for c in conns if c.conn_type == "-->"]

                for old_conn in control_conns:
                    # get from obj
                    from_obj = old_conn.from_obj

                    # add connection from fromobj to new sig
                    new_conn = Connection(from_obj, old_conn.outlet_index, sig_obj, 0, "-->")
                    self.parent_graph._PdGraph__connections.append(new_conn)
                    sig_obj.add_connection(new_conn)
                    from_obj.add_connection(new_conn)

                    # remove connection from fromobj
                    self.parent_graph._PdGraph__connections.remove(old_conn)
                    from_obj.remove_connection(old_conn)
                    self.remove_connection(old_conn)

        # make sure to call parent validate_configuration()
        super().validate_configuration()

    def __repr__(self) -> str:
        return "[{0} {1}]".format(
            os.path.splitext(os.path.basename(self._PdGraph__pd_path))[0],
            " ".join(self.obj_args[1:]))



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
# Copyright (C) 2021-2023 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import hashlib
import os


def filter_max(i: int, j: int) -> int:
    """Calculate the maximum of two integers.
    """
    return max(int(i), int(j))


def filter_plugin_id(s: str) -> int:
    """ Return a unique id from patch name
        [0...32767
    """
    sh = hashlib.md5(s.encode('utf-8'))
    sd = sh.hexdigest()[:4]
    return int(sd, 16) & 0x7FFF


def filter_string_cap(s: str, li: int) -> str:
    """Returns a truncated string with ellipsis if it exceeds a certain length.
    """
    return s if (len(s) <= li) else f"{s[0:li - 3]}..."


def filter_templates(template_name: str) -> bool:
    return False if os.path.basename(template_name) in [".DS_Store"] else True


def filter_uniqueid(s: str) -> str:
    """ Return a unique id (in hexadecimal) for the Plugin interface.
    """
    sh = hashlib.md5(s.encode('utf-8'))
    sd = sh.hexdigest().upper()[0:8]
    return f"0x{sd}"


def filter_xcode_build(s: str) -> str:
    """Return a build hash suitable for use in an Xcode project file.
    """
    s = f"{s}_build"
    sh = hashlib.md5(s.encode('utf-8'))
    return sh.hexdigest().upper()[0:24]


def filter_xcode_copy(s: str) -> str:
    """Return a copyref hash suitable for use in an Xcode project file.
    """
    s = f"{s}_copy"
    sh = hashlib.md5(s.encode('utf-8'))
    return sh.hexdigest().upper()[0:24]


def filter_xcode_fileref(s: str) -> str:
    """Return a fileref hash suitable for use in an Xcode project file.
    """
    s = f"{s}_fileref"
    sh = hashlib.md5(s.encode('utf-8'))
    return sh.hexdigest().upper()[0:24]



# import datetime
import os
import shutil
import time
import jinja2
import json
from typing import List, Optional

import hvcc.core.hv2ir.HeavyLangObject as HeavyLangObject
from ..copyright import copyright_manager

from hvcc.interpreters.pd2hv.NotificationEnum import NotificationEnum
from hvcc.types.compiler import Generator, CompilerResp, CompilerNotif, CompilerMsg, ExternInfo
from hvcc.types.meta import Meta
from hvcc.types.IR import IRGraph


heavy_hash = HeavyLangObject.HeavyLangObject.get_hash
OWL_BUTTONS = ['Push', 'B1', 'B2', 'B3', 'B4', 'B5', 'B6', 'B7', 'B8']


class c2owl(Generator):
    """ Generates a OWL wrapper for a given patch.
    """

    @classmethod
    def make_jdata(cls, patch_ir: str) -> List:
        jdata = list()

        with open(patch_ir, mode="r") as f:
            ir = IRGraph(**json.load(f))

            for name, recv in ir.control.receivers.items():
                # skip __hv_init and similar
                if name.startswith("__"):
                    continue

                # If a name has been specified
                if recv.attributes.get('raw'):
                    key = recv.attributes['raw']
                    jdata.append((key, name, 'RECV', f"0x{heavy_hash(name):X}",
                                  recv.attributes['min'],
                                  recv.attributes['max'],
                                  recv.attributes['default'],
                                  key in OWL_BUTTONS))

                elif name.startswith('Channel-'):
                    key = name.split('Channel-', 1)[1]
                    jdata.append((key, name, 'RECV', f"0x{heavy_hash(name):X}",
                                  0, 1, None, key in OWL_BUTTONS))

            for _, obj in ir.objects.items():
                try:
                    if obj.type == '__send':
                        name = obj.args['name']
                        if obj.args['attributes'].get('raw'):
                            key = obj.args['attributes']['raw']
                            jdata.append((key, f'{name}>', 'SEND', f"0x{heavy_hash(name):X}",
                                          obj.args['attributes']['min'],
                                          obj.args['attributes']['max'],
                                          obj.args['attributes']['default'],
                                          key in OWL_BUTTONS))
                        elif name.startswith('Channel-'):
                            key = name.split('Channel-', 1)[1]
                            jdata.append((key, f'{name}>', 'SEND', f"0x{heavy_hash(name):X}",
                                          0, 1, None, key in OWL_BUTTONS))
                except Exception:
                    pass

            return jdata

    @classmethod
    def compile(
        cls,
        c_src_dir: str,
        out_dir: str,
        externs: ExternInfo,
        patch_name: Optional[str] = None,
        patch_meta: Meta = Meta(),
        num_input_channels: int = 0,
        num_output_channels: int = 0,
        copyright: Optional[str] = None,
        verbose: Optional[bool] = False
    ) -> CompilerResp:

        tick = time.time()

        out_dir = os.path.join(out_dir, "Source")
        patch_name = patch_name or "heavy"
        copyright_c = copyright_manager.get_copyright_for_c(copyright)

        try:
            # ensure that the output directory does not exist
            out_dir = os.path.abspath(out_dir)
            if os.path.exists(out_dir):
                shutil.rmtree(out_dir)

            # copy over generated C source files
            shutil.copytree(c_src_dir, out_dir)

            # copy over deps
            shutil.copytree(os.path.join(os.path.dirname(__file__), "deps"), out_dir, dirs_exist_ok=True)

            # initialize the jinja template environment
            env = jinja2.Environment()

            env.loader = jinja2.FileSystemLoader(
                os.path.join(os.path.dirname(os.path.abspath(__file__)), "templates"))

            # construct jdata from ir
            ir_dir = os.path.join(c_src_dir, "../ir")
            patch_ir = os.path.join(ir_dir, f"{patch_name}.heavy.ir.json")
            jdata = cls.make_jdata(patch_ir)

            # generate OWL wrapper from template
            owl_hpp_path = os.path.join(out_dir, f"HeavyOWL_{patch_name}.hpp")
            with open(owl_hpp_path, "w") as f:
                f.write(env.get_template("HeavyOwl.hpp").render(
                    jdata=jdata,
                    name=patch_name,
                    copyright=copyright_c))
            owl_h_path = os.path.join(out_dir, "HeavyOwlConstants.h")
            with open(owl_h_path, "w") as f:
                f.write(env.get_template("HeavyOwlConstants.h").render(
                    jdata=jdata,
                    copyright=copyright_c))

            # ======================================================================================

            return CompilerResp(
                stage="c2owl",
                in_dir=c_src_dir,
                out_dir=out_dir,
                out_file=os.path.basename(owl_h_path),
                compile_time=time.time() - tick
            )

        except Exception as e:
            return CompilerResp(
                stage="c2owl",
                notifs=CompilerNotif(
                    has_error=True,
                    exception=e,
                    warnings=[],
                    errors=[CompilerMsg(
                        enum=NotificationEnum.ERROR_EXCEPTION,
                        message=str(e)
                    )]
                ),
                in_dir=c_src_dir,
                out_dir=out_dir,
                compile_time=time.time() - tick
            )



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
# Copyright (C) 2021-2024 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import subprocess
import time
import jinja2

from shutil import which
from typing import Optional

from hvcc.core.hv2ir.HeavyException import HeavyException
from ..copyright import copyright_manager

from hvcc.interpreters.pd2hv.NotificationEnum import NotificationEnum
from hvcc.types.compiler import Generator, CompilerResp, CompilerNotif, CompilerMsg, ExternInfo
from hvcc.types.meta import Meta


class c2js(Generator):
    """Compiles a directory of C source files into javascript. Requires the
    emscripten library to be installed - https://github.com/kripken/emscripten
    """

    __HV_API = [
        "_hv_{0}_new",
        "_hv_{0}_new_with_options",
        "_hv_delete",
        "_hv_processInline",
        "_hv_getNumInputChannels",
        "_hv_getNumOutputChannels",
        "_hv_samplesToMilliseconds",
        "_hv_setPrintHook",
        "_hv_setSendHook",
        "_hv_sendFloatToReceiver",
        "_hv_sendBangToReceiver",
        "_hv_sendSymbolToReceiver",
        "_hv_stringToHash",
        "_hv_msg_getByteSize",
        "_hv_msg_init",
        "_hv_msg_hasFormat",
        "_hv_msg_setFloat",
        "_hv_msg_getFloat",
        "_hv_msg_getTimestamp",
        "_hv_table_getLength",
        "_hv_table_setLength",
        "_hv_table_getBuffer",
        "_hv_sendMessageToReceiverV",
        "_hv_sendMessageToReceiverFF",
        "_hv_sendMessageToReceiverFFF",
        "_malloc"  # Rationale: https://github.com/emscripten-core/emscripten/issues/6882#issuecomment-406745898
    ]

    @classmethod
    def run_emscripten(
        cls,
        c_src_dir: str,
        out_dir: str,
        patch_name: str,
        output_name: str,
        post_js_path: str,
        should_modularize: int,
        environment: str,
        pre_js_path: str = "",
        binaryen_async: int = 1
    ) -> str:
        """Run the emcc command to compile C source files to a javascript library.
        """

        if os.name == 'nt':
            emcc_path = which("emcc.bat")
        else:
            emcc_path = which("emcc")

        if emcc_path is None:
            raise HeavyException("emcc is not in the PATH")

        c_flags = [
            f"-I {c_src_dir}",
            "-DHV_SIMD_NONE",
            "-ffast-math",
            "-DNDEBUG",
            "-Wall"
        ]

        c_src_paths = [os.path.join(c_src_dir, c) for c in os.listdir(c_src_dir) if c.endswith((".c"))]
        cpp_src_paths = [os.path.join(c_src_dir, cpp) for cpp in os.listdir(c_src_dir) if cpp.endswith((".cpp"))]
        obj_paths = []
        cmd = ""

        # compile C files
        for c in c_src_paths:
            obj_path = f"{os.path.splitext(c)[0]}.o"
            cmd = [emcc_path] + c_flags + ["-c", "-o", obj_path, c]  # type: ignore
            subprocess.check_output(cmd)  # run emscripten
            obj_paths += [obj_path]

        # compile C++ files
        for cpp in cpp_src_paths:
            obj_path = f"{os.path.splitext(cpp)[0]}.o"
            cmd = [emcc_path] + c_flags + ["-std=c++11"] + ["-c", "-o", obj_path, cpp]  # type: ignore
            subprocess.check_output(cmd)  # run emscripten
            obj_paths += [obj_path]

        # exported heavy api methods
        hv_api_defs = ", ".join([f"\"{x.format(patch_name)}\"" for x in cls.__HV_API])

        # output path
        wasm_js_path = os.path.join(out_dir, f"{output_name}.js")

        linker_flags = [
            "-O3",
            "-s", "RESERVED_FUNCTION_POINTERS=2",
            "-s", "DEFAULT_LIBRARY_FUNCS_TO_INCLUDE=$addFunction",
            "-s", f"EXPORTED_FUNCTIONS=[{hv_api_defs.format(patch_name)}]",
            "-s", f"MODULARIZE={should_modularize}",
            '-s', 'ASSERTIONS=1',
            '-s', f'ENVIRONMENT={environment}',
            '-s', 'SINGLE_FILE=1',
            '-s', 'ALLOW_TABLE_GROWTH=1',
            '-s', f'BINARYEN_ASYNC_COMPILATION={binaryen_async}',  # Set this to 0 for the worklet so we don't
                                                                   # wait for promises when instantiating
            "--post-js", post_js_path
        ]

        if len(pre_js_path):
            linker_flags = linker_flags + [
                "--pre-js", pre_js_path
            ]

        # include C/C++ obj files in js library
        cmd = [emcc_path] + obj_paths + linker_flags  # type: ignore
        subprocess.check_output(  # WASM
            cmd + [  # type: ignore
                "-s", "WASM=1",
                "-s", f"EXPORT_NAME='{output_name}_Module'",
                "-o", wasm_js_path
            ])

        # clean up
        for o in obj_paths:
            os.remove(o)

        return wasm_js_path

    @classmethod
    def compile(
        cls,
        c_src_dir: str,
        out_dir: str,
        externs: ExternInfo,
        patch_name: Optional[str] = None,
        patch_meta: Meta = Meta(),
        num_input_channels: int = 0,
        num_output_channels: int = 0,
        copyright: Optional[str] = None,
        verbose: Optional[bool] = False
    ) -> CompilerResp:

        tick = time.time()

        parameter_list = externs.parameters.inParam
        parameter_out_list = externs.parameters.outParam
        event_list = externs.events.inEvent
        event_out_list = externs.events.outEvent

        midi_list = externs.midi.inMidi
        midi_out_list = externs.midi.outMidi

        out_dir = os.path.join(out_dir, "js")
        patch_name = patch_name or "heavy"

        copyright_js = copyright_manager.get_copyright_for_c(copyright)
        copyright_html = copyright_manager.get_copyright_for_xml(copyright)

        if not os.path.exists(out_dir):
            os.makedirs(out_dir)
        out_dir = os.path.abspath(out_dir)

        try:
            # initialise the jinja template environment
            env = jinja2.Environment()
            env.loader = jinja2.FileSystemLoader(os.path.join(
                os.path.dirname(__file__),
                "template"))

            # generate heavy js wrapper from template
            # Note: this file will be incorporated into the emscripten output
            # and removed afterwards
            post_js_path = os.path.join(out_dir, "hv_wrapper.js")
            with open(post_js_path, "w") as f:
                f.write(env.get_template("hv_wrapper.js").render(
                    name=patch_name,
                    copyright=copyright_js,
                    externs=externs,
                    pool_sizes_kb=externs.memoryPoolSizesKb))

            js_path = cls.run_emscripten(c_src_dir=c_src_dir,
                                         out_dir=out_dir,
                                         patch_name=patch_name,
                                         output_name=patch_name,
                                         post_js_path=post_js_path,
                                         should_modularize=1,
                                         environment="web")

            # delete temporary files
            os.remove(post_js_path)

            js_out_file = os.path.basename(js_path)

            # generate index.html from template
            with open(os.path.join(out_dir, "index.html"), "w") as f:
                f.write(env.get_template("index.html").render(
                    name=patch_name,
                    includes=[f"./{js_out_file}"],
                    parameters=parameter_list,
                    parameters_out=parameter_out_list,
                    events=event_list,
                    events_out=event_out_list,
                    midi=midi_list,
                    midi_out=midi_out_list,
                    copyright=copyright_html))

            # generate heavy js worklet from template
            # Note: this file will be incorporated into the emscripten output
            # and removed afterwards
            post_js_path = os.path.join(out_dir, "hv_worklet.js")
            with open(post_js_path, "w") as f:
                f.write(env.get_template("hv_worklet.js").render(
                    name=patch_name,
                    copyright=copyright_js,
                    externs=externs,
                    pool_sizes_kb=externs.memoryPoolSizesKb))

            pre_js_path = os.path.join(out_dir, "hv_worklet_start.js")
            with open(pre_js_path, "w") as f:
                f.write(env.get_template("hv_worklet_start.js").render(
                    name=patch_name,
                    copyright=copyright_js,
                    externs=externs,
                    pool_sizes_kb=externs.memoryPoolSizesKb))

            js_path = cls.run_emscripten(c_src_dir=c_src_dir,
                                         out_dir=out_dir,
                                         patch_name=patch_name,
                                         output_name=f"{patch_name}_AudioLibWorklet",
                                         post_js_path=post_js_path,
                                         should_modularize=0,
                                         environment="shell,worker",
                                         pre_js_path=pre_js_path,
                                         binaryen_async=0)

            # delete temporary files
            os.remove(post_js_path)
            os.remove(pre_js_path)

            return CompilerResp(
                stage="c2js",
                in_dir=c_src_dir,
                out_dir=out_dir,
                out_file=js_out_file,
                compile_time=time.time() - tick
            )

        except Exception as e:
            return CompilerResp(
                stage="c2js",
                notifs=CompilerNotif(
                    has_error=True,
                    exception=e,
                    warnings=[],
                    errors=[CompilerMsg(
                        enum=NotificationEnum.ERROR_EXCEPTION,
                        message=str(e)
                    )]
                ),
                in_dir=c_src_dir,
                out_dir=out_dir,
                compile_time=time.time() - tick
            )



# Heavy Compiler Collection
# Copyright (C) 2024 Wasted Audio
#
# SPDX-License-Identifier: GPL-3.0-only

from typing import Dict, Literal, List, Optional, Tuple, Union
from pydantic import BaseModel, HttpUrl


DaisyBoards = Literal['pod', 'petal', 'patch', 'patch_init', 'field']
DaisyBootloader = Literal['BOOT_NONE', 'BOOT_SRAM', 'BOOT_QSPI']
DPFFormats = Literal['lv2', 'lv2_sep', 'vst2', 'vst3', 'clap', 'jack', 'dssi', 'ladspa']


class Daisy(BaseModel):
    board:          DaisyBoards = 'pod'
    board_file:     Optional[str] = None
    libdaisy_path:  Union[str, int] = 2
    usb_midi:       bool = False
    debug_printing: bool = False
    samplerate:     int = 48000
    blocksize:      Optional[int] = None
    linker_script:  str = ''
    bootloader:     Optional[DaisyBootloader] = None


class DPFUISize(BaseModel):
    width: int
    height: int


class DPFPortGroups(BaseModel):
    input: Dict[str, Dict[str, Union[int, Tuple[int, bool]]]] = {}
    output: Dict[str, Dict[str, Union[int, Tuple[int, bool]]]] = {}


class DPF(BaseModel):
    dpf_path:       str = ""
    description:    Optional[str] = None
    makefile_dep:   List[str] = []
    enable_ui:      bool = False
    enable_modgui:  bool = False
    ui_size:        Optional[DPFUISize] = None
    midi_input:     bool = False
    midi_output:    bool = False
    port_groups:    Optional[DPFPortGroups] = None
    enumerators:    Optional[Dict[str, List[str]]] = None
    denormals:      bool = True
    version:        Optional[str] = None
    license:        Optional[str] = None
    maker:          Optional[str] = None
    homepage:       Optional[HttpUrl] = None
    plugin_uri:     Optional[str] = None
    plugin_clap_id: Optional[str] = None
    plugin_formats: List[DPFFormats] = []
    lv2_info:       Optional[str] = None
    vst3_info:      Optional[str] = None
    clap_info:      List[str] = []


class Meta(BaseModel):
    name: Optional[str] = None
    nosimd: Optional[bool] = False
    daisy: Daisy = Daisy()
    dpf: DPF = DPF()



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


# moved to HeavyParser.py because of circular dependency


# from .HeavyException import HeavyException
# from .HeavyIrObject import HeavyIrObject
# from .HeavyLangObject import HeavyLangObject


# class HLangIf(HeavyLangObject):
#     """ Translates HeavyLang object [if] to HeavyIR [if] or [if~].
#     """

#     def __init__(self, obj_type, args, graph, annotations=None):
#         HeavyLangObject.__init__(self, "if", args, graph,
#                                  num_inlets=2,
#                                  num_outlets=2,
#                                  annotations=annotations)

#     def reduce(self):
#         if self.has_inlet_connection_format(["cc", "_c", "c_", "__"]):
#             x = HeavyIrObject("__if", self.args)
#         elif self.has_inlet_connection_format("ff"):
#             # TODO(mhroth): implement this
#             x = HeavyParser.graph_from_file("./hvlib/if~f.hv.json")
#         elif self.has_inlet_connection_format("ii"):
#             # TODO(mhroth): implement this
#             x = HeavyParser.graph_from_file("./hvlib/if~i.hv.json")
#         else:
#             fmt = self._get_connection_format(self.inlet_connections)
#             raise HeavyException(f"Unhandled connection configuration to object [if]: {fmt}")

#         return ({x}, self.get_connection_move_list(x))



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
# Copyright (C) 2023 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re
from typing import Optional, List, Dict

from .NotificationEnum import NotificationEnum
from .PdObject import PdObject


class PdMessageObject(PdObject):

    # only allow dollar argumnets if they are alone
    __RE_DOLLAR = re.compile(r"\$(\d+)")

    def __init__(
        self,
        obj_type: str,
        obj_args: Optional[List] = None,
        pos_x: int = 0,
        pos_y: int = 0
    ) -> None:
        assert obj_type == "msg"
        super().__init__("msg", obj_args, pos_x, pos_y)

        self.obj_dict: Dict = {}
        semi_split: List = []

        # parse messages
        # remove prepended slash from $. Heavy does not use that.
        if obj_args is not None:
            semi_split = obj_args[0].replace(r"\$", "$").split(r"\;")
            semi_split = [x for x in semi_split if x]  # remove empty strings

        # parse local messages
        # ensure that empty message are not passed on
        if len(semi_split) > 0:
            self.obj_dict["local"] = [li.strip().split() for li in semi_split[0].split(r"\,") if len(li.strip()) > 0]
        else:
            self.obj_dict["local"] = []
            self.add_warning(
                "Message object is empty. Can it be removed?",
                NotificationEnum.WARNING_EMPTY_MESSAGE)

        # heavy does not support messages such as "$1-$2"
        for li in self.obj_dict["local"]:
            for m in li:
                x = self.__RE_DOLLAR.search(m)
                if x and len(x.group(0)) < len(m):
                    self.add_error(
                        "Heavy does not yet support message concatenation. "
                        "Dollar arguments must be alone: " + m)

        # parse remote messages
        self.obj_dict["remote"] = []
        for li in semi_split[1:]:
            l_split = li.strip().split()
            self.obj_dict["remote"].append({
                "receiver": l_split[0],
                "message": l_split[1:]
            })

    def to_hv(self) -> Dict:
        return {
            "type": "message",
            "args": self.obj_dict,
            "properties": {
                "x": self.pos_x,
                "y": self.pos_y
            }
        }



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
# Copyright (C) 2023-2024 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Callable, Dict, List

from .HeavyObject import HeavyObject

from hvcc.types.IR import IRObjectdict, IROnMessage


class ControlSend(HeavyObject):

    c_struct = "ControlSend"
    preamble = "cSend"

    @classmethod
    def get_C_onMessage(cls, obj_type: str, obj_id: str, inlet_index: int, args: Dict) -> List[str]:
        return [f"cSend_{obj_id}_sendMessage(_c, 0, m);"]

    @classmethod
    def get_C_impl(
        cls,
        obj_type: str,
        obj_id: str,
        on_message_list: List[List[IROnMessage]],
        get_obj_class: Callable,
        objects: Dict[str, IRObjectdict]
    ) -> List[str]:
        # Note(joe): if no corresponding receivers exist and there's no extern indicator
        # then there is not much need to generate code stub
        send_message_list = [
            "{0}_{1}_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *m) {{".format(
                cls.get_preamble(obj_type),
                obj_id)]

        if objects[obj_id].args.get("extern", False):
            # call the send hook
            send_name = objects[obj_id].args["name"]
            send_message_list.append("if (_c->getSendHook() != nullptr) _c->getSendHook()(_c, \"{0}\", {1}, m);".format(
                send_name,
                cls.get_hash_string(send_name)))

        # a send has only one (implicit!) outlet
        send_message_list.extend(
            cls._get_on_message_list(on_message_list[0], get_obj_class, objects))

        send_message_list.append("}")  # end function

        return send_message_list



# Heavy Compiler Collection
# Copyright (C) 2024 Wasted Audio
#
# SPDX-License-Identifier: GPL-3.0-only

from pydantic import BaseModel, RootModel
from typing import Dict, List, Optional, Union

from hvcc.types.Lang import LangLetType
from hvcc.version import VERSION


# IR Object Definitions

IRValue = Union[float, int, str, List[float], List[int]]


class IRArg(BaseModel):
    name: str
    value_type: str
    description: str = ""
    default: Optional[IRValue] = None
    required: bool


class IR(BaseModel):
    control: bool
    signal: bool
    init: bool


class Perf(BaseModel):
    avx: float = 0
    sse: float = 0
    neon: float = 0


class IRNode(BaseModel):
    inlets: List[LangLetType]
    ir: IR
    outlets: List[LangLetType]
    args: List[IRArg] = []
    perf: Optional[Perf] = Perf()
    # perf: Perf
    description: Optional[str] = None
    alias: List[str] = []
    tags: List[str] = []
    keywords: List[str] = []


class HeavyIRType(RootModel):
    root: Dict[str, IRNode]


# IR Graph Definitions

class IRName(BaseModel):
    escaped: str
    display: str


class IRObjectdict(BaseModel):
    args: Dict
    type: str


class IRInit(BaseModel):
    order: List[str] = []


class IRReceiver(BaseModel):
    display: str
    hash: str
    extern: Optional[str] = None
    attributes: Dict
    ids: List[str]


class IROnMessage(BaseModel):
    id: str
    inletIndex: Optional[int] = None


class IRSendMessage(BaseModel):
    """ This object is a mess and should be split up
    """
    id: str
    onMessage: List[List[IROnMessage]]  # TODO: why is it a nested List?
    type: Optional[str] = None
    extern: Optional[str] = ''
    attributes: Optional[Union[List, Dict]] = None  # TODO: List from IR and Dict from Lang
    hash: Optional[str] = None
    display: Optional[str] = None
    name: str = ''


class IRTable(BaseModel):
    id: str
    display: str
    hash: str
    extern: Optional[bool] = None


class IRControl(BaseModel):
    receivers: Dict[str, IRReceiver]
    sendMessage: List[IRSendMessage]


class IRNumTempBuffer(BaseModel):
    float: int
    integer: int


class IRBuffer(BaseModel):
    type: str
    index: int


class IRSignalList(BaseModel):
    id: str
    inputBuffers: List[IRBuffer]
    outputBuffers: List[IRBuffer]


class IRSignal(BaseModel):
    numInputBuffers: int
    numOutputBuffers: int
    numTemporaryBuffers: IRNumTempBuffer
    processOrder: List[IRSignalList]


class IRGraph(BaseModel):
    version: str = VERSION
    name: IRName
    objects: Dict[str, IRObjectdict] = {}
    init: IRInit
    tables: Dict[str, IRTable] = {}
    control: IRControl
    signal: IRSignal


if __name__ == "__main__":
    """ Test object definitions
    """
    import json
    import importlib_resources

    heavy_ir_json = importlib_resources.files('hvcc') / 'core/json/heavy.ir.json'
    with open(heavy_ir_json, "r") as f:
        data = json.load(f)
        heavy_ir = HeavyIRType(root=data)
        print(heavy_ir.root.keys())



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
# Copyright (C) 2023-2024 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re
from typing import Optional, Dict

from .HeavyLangObject import HeavyLangObject
from .HeavyIrObject import HeavyIrObject
from .HeavyGraph import HeavyGraph


class HLangTable(HeavyLangObject):
    """ Handles the HeavyLang "table" object.
    """

    def __init__(
        self,
        obj_type: str,
        args: Dict,
        graph: 'HeavyGraph',
        annotations: Optional[Dict] = None
    ) -> None:
        assert obj_type == "table"
        super().__init__(obj_type, args, graph, annotations=annotations)

        # the values argument overrides the size argument
        if len(self.args.get("values", [])) > 0:
            self.args["size"] = len(self.args["values"])

        if self.args["extern"]:
            # externed tables must contain only alphanumeric characters or underscores,
            # so that the names can be easily and transparently turned into code
            if re.search(r"\W", args["name"]):
                self.add_error(f"Table names may only contain alphanumeric characters \
                                or underscore: '{args['name']}'")

    def reduce(self) -> tuple:
        x = HeavyIrObject("__table", self.args)
        # ensure that __table object maintains the same id as the original
        # table object. The latter is referenced by id from other objects.
        # Consistency must be maintained.
        x.id = self.id
        return ({x}, self.get_connection_move_list(x))



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
# Copyright (C) 2023 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Optional, List, Dict

from .NotificationEnum import NotificationEnum
from .PdObject import PdObject


class PdPackObject(PdObject):
    def __init__(
        self,
        obj_type: str,
        obj_args: Optional[List] = None,
        pos_x: int = 0,
        pos_y: int = 0
    ) -> None:
        assert obj_type == "pack"
        super().__init__(obj_type, obj_args, pos_x, pos_y)

        self.values = []

        if len(self.obj_args) == 0:
            self.values = [0.0, 0.0]

        for x in self.obj_args:
            try:
                self.values.append(float(x))
            except Exception:
                if x in {"f", "float"}:
                    self.values.append(0.0)
                else:
                    self.add_error(
                        f"\"{x}\" argument to [pack] object not supported.",
                        NotificationEnum.ERROR_PACK_FLOAT_ARGUMENTS)

    def to_hv(self) -> Dict:
        return {
            "type": "__pack",
            "args": {
                "values": self.values
            },
            "properties": {
                "x": self.pos_x,
                "y": self.pos_y
            }
        }



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
# Copyright (C) 2021-2024 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import shutil
import time
import jinja2
from typing import Optional

from ..copyright import copyright_manager
from ..filters import filter_max

from hvcc.interpreters.pd2hv.NotificationEnum import NotificationEnum
from hvcc.types.compiler import Generator, CompilerResp, CompilerNotif, CompilerMsg, ExternInfo
from hvcc.types.meta import Meta


class c2pdext(Generator):
    """Generates a Pure Data external wrapper for a given patch.
    """

    @classmethod
    def compile(
        cls,
        c_src_dir: str,
        out_dir: str,
        externs: ExternInfo,
        patch_name: Optional[str] = None,
        patch_meta: Meta = Meta(),
        num_input_channels: int = 0,
        num_output_channels: int = 0,
        copyright: Optional[str] = None,
        verbose: Optional[bool] = False
    ) -> CompilerResp:

        tick = time.time()

        out_dir = os.path.join(out_dir, "pdext")
        receiver_list = externs.parameters.inParam

        copyright = copyright_manager.get_copyright_for_c(copyright)

        patch_name = patch_name or "heavy"
        ext_name = f"{patch_name}~"
        struct_name = f"{patch_name}_tilde"

        # ensure that the output directory does not exist
        out_dir = os.path.abspath(out_dir)
        if os.path.exists(out_dir):
            shutil.rmtree(out_dir)

        # copy over generated C source files
        shutil.copytree(c_src_dir, out_dir)

        # copy over static files
        shutil.copy(
            os.path.join(os.path.dirname(os.path.abspath(__file__)), "static", "m_pd.h"),
            f"{out_dir}/")
        shutil.copy(
            os.path.join(os.path.dirname(os.path.abspath(__file__)), "static", "Makefile.pdlibbuilder"),
            f"{out_dir}/../")

        try:
            # initialise the jinja template environment
            env = jinja2.Environment()
            env.filters["max"] = filter_max
            env.loader = jinja2.FileSystemLoader(
                os.path.join(os.path.dirname(os.path.abspath(__file__)), "templates"))

            # generate Pd external wrapper from template
            pdext_path = os.path.join(out_dir, f"{ext_name}.c")
            with open(pdext_path, "w") as f:
                f.write(env.get_template("pd_external.c").render(
                    name=patch_name,
                    struct_name=struct_name,
                    display_name=ext_name,
                    num_input_channels=num_input_channels,
                    num_output_channels=num_output_channels,
                    receivers=receiver_list,
                    copyright=copyright))

            # generate Makefile from template
            pdext_path = os.path.join(out_dir, "../Makefile")
            with open(pdext_path, "w") as f:
                f.write(env.get_template("Makefile").render(
                    name=patch_name))

            return CompilerResp(
                stage="c2pdext",
                in_dir=c_src_dir,
                out_dir=out_dir,
                out_file=os.path.basename(pdext_path),
                compile_time=time.time() - tick
            )

        except Exception as e:
            return CompilerResp(
                stage="c2pdext",
                notifs=CompilerNotif(
                    has_error=True,
                    exception=e,
                    warnings=[],
                    errors=[CompilerMsg(
                        enum=NotificationEnum.ERROR_EXCEPTION,
                        message=str(e)
                    )]
                ),
                in_dir=c_src_dir,
                out_dir=out_dir,
                compile_time=time.time() - tick
            )



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
# Copyright (C) 2023-2024 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import os

from typing import Dict, List, Optional, TYPE_CHECKING

from .Connection import Connection
from .HeavyException import HeavyException
from .HeavyLangObject import HeavyLangObject
from .BufferPool import BufferPool

from hvcc.types.IR import HeavyIRType, IRNode, IRObjectdict, IRSendMessage, IROnMessage, IRSignalList, IRBuffer
from hvcc.types.Lang import LangLetType


if TYPE_CHECKING:
    from .HeavyGraph import HeavyGraph


class HeavyIrObject(HeavyLangObject):
    """ Intermediate Representation (IR) objects are atomic and have
        strictly defined interfaces and types. These are generally defined in
        the file heavy.ir.json.
    """

    # load the HeavyIR object definitions
    with open(os.path.join(os.path.dirname(__file__), "../json/heavy.ir.json"), "r") as f:
        __HEAVY_OBJS_IR_DICT = HeavyIRType(**json.load(f)).root

    def __init__(
        self,
        obj_type: str,
        args: Optional[Dict] = None,
        graph: Optional['HeavyGraph'] = None,
        num_inlets: int = -1,
        num_outlets: int = -1,
        annotations: Optional[Dict] = None
    ) -> None:
        # allow the number of inlets and outlets to be overridden
        num_inlets = (len(self.__HEAVY_OBJS_IR_DICT[obj_type].inlets)
                      if num_inlets < 0
                      else num_inlets)

        num_outlets = (len(self.__HEAVY_OBJS_IR_DICT[obj_type].outlets)
                       if num_outlets < 0
                       else num_outlets)

        super().__init__(obj_type, args, graph, num_inlets, num_outlets, annotations)

        # resolve arguments and fill in missing defaults for HeavyIR objects
        self.__resolve_default_ir_args()

        # the list of signal buffers at the inlets and outlets
        # these are filled in by HeavyGraph.assign_signal_buffers()
        self.inlet_buffers = [("zero", 0)] * self.num_inlets
        self.outlet_buffers = [("zero", 0)] * self.num_outlets

        # True if this object has already been ordered in the signal chain
        self.__is_ordered = False

    def __resolve_default_ir_args(self) -> None:
        """ Resolves missing default arguments. Also checks to make sure that all
            required arguments are present.
        """
        if self.type in self.__HEAVY_OBJS_IR_DICT.keys():
            for arg in self.__obj_desc.args:
                if arg.name not in self.args:
                    # if a defined argument is not in the argument dictionary
                    if not arg.required:
                        # if the argument is not required, use the default
                        self.args[arg.name] = arg.default
                    else:
                        self.add_error(f"Required argument \"{arg.name}\" not present for object {self}.")
                else:
                    # enforce argument types.
                    # if the default argument is null, don't worry about about the arg
                    if arg.default is not None:
                        self.args[arg.name] = self.force_arg_type(
                            self.args[arg.name],
                            arg.value_type,
                            self.graph)

    @classmethod
    def is_ir(cls, obj_type: str) -> bool:
        """Returns true if the type is an IR object. False otherwise.
        """
        return obj_type in cls.__HEAVY_OBJS_IR_DICT.keys()

    @property
    def does_process_signal(self) -> bool:
        """Returns True if this object processes a signal. False otherwise.
        """
        return self.__obj_desc.ir.signal

    @property
    def __obj_desc(self) -> IRNode:
        """ Returns the original HeavyIR object description.
        """
        return self.__HEAVY_OBJS_IR_DICT[self.type]

    def inlet_requires_signal(self, inlet_index: int = 0) -> bool:
        """ Returns True if the indexed inlet requires a signal connection. False otherwise.
        """
        return self.__obj_desc.inlets[inlet_index] in {"~i>", "~f>"}

    def outlet_requires_signal(self, inlet_index: int = 0) -> bool:
        """ Returns True if the indexed outlet requires a signal connection. False otherwise.
        """
        return self.__obj_desc.outlets[inlet_index] in {"~i>", "~f>"}

    def reduce(self) -> Optional[tuple]:
        # A Heavy IR object is already reduced. Returns itself and no connection changes.
        return ({self}, [])

    def get_parent_order(self) -> List:
        """ Returns a list of all objects in process order, with this object at the end.
        """
        if self.__is_ordered:
            return []
        else:
            self.__is_ordered = True
            if self.is_root():
                return [self]
            else:
                order_list = []
                for c in [c for inlet in self.inlet_connections for c in inlet]:
                    order_list.extend(c.from_object.get_parent_order())
                order_list.append(self)
                return order_list

    def assign_signal_buffers(self, buffer_pool: Optional[BufferPool]) -> None:
        if buffer_pool is not None:
            # assign the inlet buffers
            for cc in self.inlet_connections:
                cc = [c for c in cc if c.is_signal]  # only need to deal with signal connections
                if len(cc) == 0:
                    continue
                if len(cc) == 1:
                    c = cc[0]  # get the connection

                    # get the buffer at the outlet of the connected object
                    buf = c.from_object.outlet_buffers[c.outlet_index]

                    # assign the buffer to the inlet of this object
                    self.inlet_buffers[c.inlet_index] = buf

                    # decrease the retain count of the buffer
                    buffer_pool.release_buffer(buf)
                else:
                    raise HeavyException(f"This object has {len(cc)} (> 1) signal inputs.")

            # assign the output buffers
            exclude_set: set = set()

            for i in range(self.num_outlets):
                # buffers are assigned even if the outlet has no connections.
                # The buffer will still be filled. However, if the buffer has already
                # been set (i.e. non-zero) (e.g. in the case of dac~),
                # then we skip this set

                connection_type = self._resolved_outlet_type(outlet_index=i)
                if Connection.is_signal_type(connection_type) and self.outlet_buffers[i][0] == "zero" \
                        and connection_type is not None:
                    b = buffer_pool.get_buffer(
                        connection_type,
                        len(self.outlet_connections[i]),
                        exclude_set)
                    self.outlet_buffers[i] = b

                    # if the buffer has no dependencies, make sure that it isn't reused
                    # right away. All outlets should have independent buffers.
                    if len(self.outlet_connections[i]) == 0:
                        exclude_set.add(b)

    def _resolved_outlet_type(self, outlet_index: int = 0) -> Optional[LangLetType]:
        """ Returns the connection type at the given outlet.
            This information is always well-defined for IR objects.
        """
        return self.__obj_desc.outlets[outlet_index]

    #
    # Intermediate Representation generators
    #

    def get_object_dict(self) -> Dict[str, IRObjectdict]:
        """ Returns a dictionary of all constituent low-level objects,
            indexed by id, including their arguments and type.
        """
        return {
            self.id: IRObjectdict(
                args=self.args,
                type=self.type
            )
        }

    def get_ir_init_list(self) -> List[str]:
        """ Returns a list of all object id for obejcts that need initialisation.
        """
        return [self.id] if self.__obj_desc.ir.init else []

    def get_ir_on_message(self, inlet_index: int = 0) -> List[IROnMessage]:
        """ Returns an array of dictionaries containing the information for the
            corresponding on_message call.
        """
        return [IROnMessage(
            id=self.id,
            inletIndex=inlet_index
        )]

    def get_ir_control_list(self) -> List[IRSendMessage]:
        """ Returns the intermediate representation for object control functions.
            Basically, does sendMessage() need to be written?
        """
        if self.__obj_desc.ir.control:
            on_message_list = []
            for connections in self.outlet_connections:
                on_messages_let = []
                # only look at control connections
                for c in [c for c in connections if c.is_control]:
                    on_messages_let.extend(c.to_object.get_ir_on_message(c.inlet_index))
                on_message_list.append(on_messages_let)

            return [IRSendMessage(
                id=self.id,
                onMessage=on_message_list
            )]
        else:
            return []

    def get_ir_signal_list(self) -> List[IRSignalList]:
        """ Returns the intermediate representation for object process functions.
            Only outputs buffer information for lets that require a signal.
        """
        # we assume that this method will only be called on signal objects
        assert self.__obj_desc.ir.signal

        return [IRSignalList(
            id=self.id,
            inputBuffers=[
                IRBuffer(type=b[0], index=b[1]) for i, b in enumerate(self.inlet_buffers)
                if self.inlet_requires_signal(i)
            ],
            outputBuffers=[
                IRBuffer(type=b[0], index=b[1]) for i, b in enumerate(self.outlet_buffers)
                if self.outlet_requires_signal(i)
            ]
        )]



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
# Copyright (C) 2021-2024 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import shutil
import time
import jinja2
from typing import Optional

from ..copyright import copyright_manager
from ..filters import filter_plugin_id

from hvcc.interpreters.pd2hv.NotificationEnum import NotificationEnum
from hvcc.types.compiler import Generator, CompilerResp, CompilerNotif, CompilerMsg, ExternInfo
from hvcc.types.meta import Meta


class c2wwise(Generator):
    """Generates a plugin wrapper for Audiokinetic's Wwise game audio middleware
    platform.
    """

    @classmethod
    def compile(
            cls,
            c_src_dir: str,
            out_dir: str,
            externs: ExternInfo,
            patch_name: Optional[str] = None,
            patch_meta: Meta = Meta(),
            num_input_channels: int = 0,
            num_output_channels: int = 0,
            copyright: Optional[str] = None,
            verbose: Optional[bool] = False
    ) -> CompilerResp:
        tick = time.time()

        in_parameter_list = externs.parameters.inParam
        out_parameter_list = externs.parameters.outParam
        event_list = externs.events.inEvent
        out_event_list = externs.events.outEvent
        table_list = externs.tables
        patch_name = patch_name or "heavy"
        copyright_c = copyright_manager.get_copyright_for_c(copyright)
        copyright_xml = copyright_manager.get_copyright_for_xml(copyright)

        templates_dir = os.path.join(os.path.dirname(__file__), "templates")
        is_source_plugin = num_input_channels == 0
        plugin_type = "Source" if is_source_plugin else "FX"
        plugin_id = filter_plugin_id(patch_name)

        out_dir = os.path.join(out_dir, "wwise")
        if not os.path.exists(out_dir):
            os.makedirs(out_dir)

        env = jinja2.Environment()
        env.loader = jinja2.FileSystemLoader(
            encoding="utf-8-sig",
            searchpath=[templates_dir])
        try:
            channel_config = None
            if num_output_channels == 0:
                # Workaround to support patches that don't have any outputs,
                # e.g. that only ever set RTPCs or post events
                channel_config = "AK_SPEAKER_SETUP_MONO"
            elif num_output_channels == 1:
                channel_config = "AK_SPEAKER_SETUP_MONO"
            elif num_output_channels == 2:
                channel_config = "AK_SPEAKER_SETUP_STEREO"
            elif num_output_channels == 6:
                channel_config = "AK_SPEAKER_SETUP_5POINT1"
            elif num_output_channels == 8:
                channel_config = "AK_SPEAKER_SETUP_7POINT1"
            elif num_output_channels == 12:
                channel_config = "AK_SPEAKER_SETUP_DOLBY_7_1_4"
            else:
                raise Exception("Wwise plugins support only mono, stereo, 5.1, 7.1, "
                                "and 7.1.4 (Atmos) configurations, number of [dac~] channels should be appropriate")

            if plugin_type == "FX":
                if num_input_channels != num_output_channels:
                    raise Exception("Wwise FX plugins require the same input/output channel"
                                    "configuration (i.e. [adc~ 1] -> [dac~ 1]).")

            patch_src_dir = os.path.join(out_dir, "SoundEnginePlugin", "Heavy")
            if os.path.exists(patch_src_dir):
                shutil.rmtree(patch_src_dir)
            shutil.copytree(c_src_dir, patch_src_dir)

            src_ext_list = ["h", "hpp", "c", "cpp", "xml", "def", "rc", "lua", "json"]
            for f in env.list_templates(extensions=src_ext_list):
                file_dir = os.path.join(out_dir, os.path.dirname(f))
                file_name = os.path.basename(f)

                file_name = file_name.replace("{{name}}", patch_name)
                file_name = file_name.replace("{{plugin_type}}", plugin_type)
                file_path = os.path.join(file_dir, file_name)

                if not os.path.exists(os.path.dirname(file_path)):
                    os.makedirs(os.path.dirname(file_path))

                with open(file_path, "w") as g:
                    g.write(env.get_template(f).render(
                        name=patch_name,
                        parameters=in_parameter_list,
                        out_params=out_parameter_list,
                        out_events=out_event_list,
                        events=event_list,
                        tables=table_list,
                        pool_sizes_kb=externs.memoryPoolSizesKb,
                        is_source=is_source_plugin,
                        num_input_channels=num_input_channels,
                        num_output_channels=num_output_channels,
                        channel_config=channel_config,
                        plugin_type=plugin_type,
                        plugin_id=plugin_id,
                        copyright=copyright_xml if file_name.endswith(".xml") else copyright_c))

            return CompilerResp(
                stage="c2wwise",
                in_dir=c_src_dir,
                out_dir=out_dir,
                compile_time=time.time() - tick
            )

        except Exception as e:
            return CompilerResp(
                stage="c2wwise",
                notifs=CompilerNotif(
                    has_error=True,
                    exception=e,
                    warnings=[],
                    errors=[CompilerMsg(
                        enum=NotificationEnum.ERROR_EXCEPTION,
                        message=str(e)
                    )]
                ),
                in_dir=c_src_dir,
                out_dir=out_dir,
                compile_time=time.time() - tick
            )



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
# Copyright (C) 2023 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Optional, List, Dict

from .Connection import Connection
from .HeavyObject import HeavyObject
from .PdObject import PdObject


class PdBinopObject(PdObject):
    # a translation dictionary from a Pd object to corresponding heavy object
    __PD_HEAVY_DICT = {
        "+": "+",
        "+~": "+",
        "-": "-",
        "-~": "-",
        "*": "*",
        "*~": "*",
        "/": "/",
        "/~": "/",
        "mod": "mod",
        "%": "%",
        "max": "max",
        "max~": "max",
        "min": "min",
        "min~": "min",
        "&": "&",
        "&&": "&&",
        "|": "|",
        "||": "||",
        "==": "==",
        "!=": "!=",
        "<": "<",
        "<=": "<=",
        ">": ">",
        ">=": ">=",
        "pow": "pow",
        "pow~": "pow",
        ">>": ">>",
        "<<": "<<"
    }

    def __init__(
        self,
        obj_type: str,
        obj_args: Optional[List] = None,
        pos_x: int = 0,
        pos_y: int = 0
    ) -> None:
        assert self.is_binop(obj_type)
        super().__init__(obj_type, obj_args, pos_x, pos_y)

    @classmethod
    def is_binop(cls, obj_type: str) -> bool:
        return obj_type in cls.__PD_HEAVY_DICT

    @classmethod
    def get_supported_objects(cls) -> set:
        return set(cls.__PD_HEAVY_DICT.keys())

    def validate_configuration(self) -> None:
        # check signal objects for control connections and auto insert
        # heavy var objects where necessary
        if self.obj_type.endswith("~"):

            # left inlet check
            conns_left = self._inlet_connections.get("0", [])

            if len(conns_left) == 0:
                # no left inlet connections
                self.convert_ctrl_to_sig_connections_at_inlet([], 0)
            elif (len([c for c in conns_left if c.conn_type == "~f>"]) == 0) and len(conns_left) > 0:
                # control connection and no sig connection
                self.convert_ctrl_to_sig_connections_at_inlet(conns_left, 0)

            # right inlet check
            conns_right = self._inlet_connections.get("1", [])
            num_signal_conns = len([c for c in conns_right if c.conn_type == "~f>"])

            if len(self.obj_args) > 0 and num_signal_conns > 0:
                # TODO(joe): removing this connection instead of throwing an error
                # would be compatible with Pd's behaviour
                self.add_error("signal outlet connected to nonsignal inlet")

            if len(self.obj_args) == 0 and num_signal_conns == 0 and len(conns_right) > 0:
                # any arguments present will create a control type right inlet
                self.convert_ctrl_to_sig_connections_at_inlet(conns_right, 1)

        if len(self.obj_args) > 0:
            try:
                self.__k = float(self.obj_args[0])
            except Exception:
                self.add_warning(f"\"{self.obj_args[0]}\" cannot be resolved to a number. Defaulting to zero.")
                self.__k = 0.0
        else:
            self.__k = 0.0

    def convert_ctrl_to_sig_connections_at_inlet(self, connection_list: List, inlet_index: int) -> None:
        """ Auto insert heavy var object inbetween control connections.
        """
        sig_obj = HeavyObject(obj_type="var",
                              obj_args=[0],
                              pos_x=int(self.pos_x),
                              pos_y=int(self.pos_y - 5))  # shift upwards a few points

        # add sig~ object to parent graph
        if self.parent_graph is not None:
            self.parent_graph.add_object(sig_obj)

            # add connection from sig~ to this object
            c = Connection(sig_obj, 0, self, inlet_index, "~f>")
            self.parent_graph._PdGraph__connections.append(c)  # update the local connections list
            sig_obj.add_connection(c)
            self.add_connection(c)

            # retrieve all control connections
            control_conns = [c for c in connection_list if c.conn_type == "-->"]

            for old_conn in control_conns:
                # get from obj
                from_obj = old_conn.from_obj

                # add connection from fromobj to new sig
                new_conn = Connection(from_obj, old_conn.outlet_index, sig_obj, 0, "-->")
                self.parent_graph._PdGraph__connections.append(new_conn)
                sig_obj.add_connection(new_conn)
                from_obj.add_connection(new_conn)

                # remove connection from fromobj
                self.parent_graph._PdGraph__connections.remove(old_conn)
                from_obj.remove_connection(old_conn)
                self.remove_connection(old_conn)

    def to_hv(self) -> Dict:
        return {
            "type": self.__PD_HEAVY_DICT[self.obj_type],
            "args": {
                "k": self.__k
            },
            "properties": {
                "x": self.pos_x,
                "y": self.pos_y
            }
        }



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
# Copyright (C) 2021-2024 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import json
import os
import sys
import time

from hvcc.version import VERSION
from hvcc.compiler import compile_dataflow


class Colours:
    purple = "\033[95m"
    cyan = "\033[96m"
    dark_cyan = "\033[36m"
    blue = "\033[94m"
    green = "\033[92m"
    yellow = "\033[93m"
    red = "\033[91m"
    bold = "\033[1m"
    underline = "\033[4m"
    end = "\033[0m"


def main() -> bool:
    tick = time.time()

    parser = argparse.ArgumentParser(
        description="This is the Wasted Audio Heavy compiler. It compiles supported dataflow languages into C,"
                    " and other supported frameworks.")
    parser.add_argument(
        "in_path",
        help="The input dataflow file.")
    parser.add_argument(
        "-o",
        "--out_dir",
        help="Build output path.")
    parser.add_argument(
        "-p",
        "--search_paths",
        nargs="+",
        help="Add a list of directories to search through for abstractions.")
    parser.add_argument(
        "-n",
        "--name",
        default="heavy",
        help="Provides a name for the generated Heavy context.")
    parser.add_argument(
        "-m",
        "--meta",
        help="Provide metadata file (json) for generator")
    parser.add_argument(
        "-g",
        "--gen",
        nargs="+",
        default=["c"],
        help="List of generator outputs: c, unity, wwise, js, pdext, daisy, dpf, fabric, owl")
    parser.add_argument(
        "-G",
        "--ext-gen",
        nargs="*",
        help="List of external generator modules, see 'External Generators' docs page.")
    parser.add_argument(
        "--results_path",
        help="Write results dictionary to the given path as a JSON-formatted string."
             " Target directory will be created if it does not exist.")
    parser.add_argument(
        "--nodsp",
        action='store_true',
        help="Disable DSP. Run as control-only patch."
    )
    parser.add_argument(
        "-v",
        "--verbose",
        help="Show debugging information.",
        action="count")
    parser.add_argument(
        "--copyright",
        help="A string indicating the owner of the copyright.")
    parser.add_argument(
        "-V",
        "--version",
        action='version',
        help="Print version and exit.",
        version=VERSION
    )
    args = parser.parse_args()

    in_path = os.path.abspath(args.in_path)
    results = compile_dataflow(
        in_path=in_path,
        out_dir=args.out_dir or os.path.dirname(in_path),
        patch_name=args.name,
        patch_meta_file=args.meta,
        search_paths=args.search_paths,
        generators=args.gen,
        ext_generators=args.ext_gen,
        verbose=args.verbose,
        copyright=args.copyright,
        nodsp=args.nodsp
    )

    errorCount = 0
    for r in list(results.root.values()):
        # print any errors
        if r.notifs.has_error:
            for i, error in enumerate(r.notifs.errors):
                errorCount += 1
                print("{4:3d}) {2}Error{3} {0}: {1}".format(
                    r.stage, error.message, Colours.red, Colours.end, i + 1))

            # only print exception if no errors are indicated
            if len(r.notifs.errors) == 0 and r.notifs.exception is not None:
                errorCount += 1
                print("{2}Error{3} {0} exception: {1}".format(
                    r.stage, r.notifs.exception, Colours.red, Colours.end))

            # clear any exceptions such that results can be JSONified if necessary
            r.notifs.exception = None

        # print any warnings
        for i, warning in enumerate(r.notifs.warnings):
            print("{4:3d}) {2}Warning{3} {0}: {1}".format(
                r.stage, warning.message, Colours.yellow, Colours.end, i + 1))

    if args.results_path:
        results_path = os.path.realpath(os.path.abspath(args.results_path))
        results_dir = os.path.dirname(results_path)

        if not os.path.exists(results_dir):
            os.makedirs(results_dir)

        with open(results_path, "w") as f:
            json.dump(results.model_dump(), f)

    if args.verbose:
        print("Total compile time: {0:.2f}ms".format(1000 * (time.time() - tick)))
    return errorCount != 0


if __name__ == "__main__":
    ret = main()
    sys.exit(ret)



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
# Copyright (C) 2023-2024 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Optional, Dict, List

from .Connection import Connection
from .HeavyException import HeavyException
from .HeavyIrObject import HeavyIrObject
from .HeavyLangObject import HeavyLangObject
from .HeavyGraph import HeavyGraph


class HLangSequence(HeavyLangObject):

    def __init__(
        self,
        obj_type: str,
        args: Dict,
        graph: 'HeavyGraph',
        annotations: Optional[Dict] = None
    ) -> None:
        # get the number of outlets that this object has
        num_outlets = len(args[self._HEAVY_LANG_DICT[obj_type].args[0].name])
        super().__init__(obj_type, args, graph,
                         num_inlets=1,
                         num_outlets=num_outlets,
                         annotations=annotations)

    def reduce(self) -> tuple:
        cast_objs: List = []
        for a in self.args[self.name_for_arg()]:
            if a in {"a", "anything", "l", "list"}:
                cast_objs.append(None)  # pass through
            elif a in {"b", "bang"}:
                cast_objs.append(HeavyIrObject("__cast_b", {}))
            elif a in {"f", "float"}:
                cast_objs.append(HeavyIrObject("__cast_f", {}))
            elif a in {"s", "symbol"}:
                cast_objs.append(HeavyIrObject("__cast_s", {}))
            else:
                raise HeavyException(f"Unsupported cast type '{a}'.")

        connections = []

        # establish connections from inlet objects to the cast objects
        for ci in self.inlet_connections[0]:
            c_list = []
            for i in range(self.num_outlets - 1, -1, -1):
                if cast_objs[i] is not None:
                    c_list.append(Connection.copy(ci, to_object=cast_objs[i], inlet_index=0))
                else:
                    # for anything casts, establish a connection directly to the target
                    for co in self.outlet_connections[i]:
                        c_list.append(Connection.copy(ci, to_object=co.to_object, inlet_index=co.inlet_index))
            connections.append((ci, c_list))

        # establish connections from the cast objects
        for i in range(self.num_outlets - 1, -1, -1):
            if cast_objs[i] is not None:
                for c in self.outlet_connections[i]:
                    connections.append((c, [c.copy(from_object=cast_objs[i], outlet_index=0)]))

        # remove None objects from cast_objs. It was only used as a placeholder
        # to indicate that a cast passthrough
        objects = set(cast_objs)
        objects.discard(None)

        return (objects, connections)



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


# moved to HeavyParser.py because of circular dependency


# import os
# import random

# from .HeavyLangObject import HeavyLangObject
# from .HeavyParser import HeavyParser


# class HLangNoise(HeavyLangObject):
#     """ Handles the HeavyLang "noise" object.
#     """

#     def __init__(self, obj_type, args, graph, annotations=None):
#         assert obj_type == "noise"
#         HeavyLangObject.__init__(self, "noise", args, graph,
#                                  num_inlets=1,
#                                  num_outlets=1,
#                                  annotations=annotations)

#     def reduce(self):
#         seed = int(random.uniform(1, 2147483647))  # assign a random 32-bit seed
#         noise_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), "./hvlib/noise.hv.json")
#         x = HeavyParser.graph_from_file(noise_path, graph_args={"seed": seed})
#         x.reduce()
#         # TODO(mhroth): deal with control input
#         return ({x}, self.get_connection_move_list(x))



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
# Copyright (C) 2023 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Optional, Dict

from .HeavyLangObject import HeavyLangObject
from .HIrSend import HIrSend
from .HeavyGraph import HeavyGraph


class HLangSend(HeavyLangObject):
    """ Translates HeavyLang object [send] to HeavyIR [__send] if control,
        otherwise for signal connections it will remove the send~/receive~ objects
        and reorder the graph.
    """

    def __init__(
        self,
        obj_type: str,
        args: Dict,
        graph: 'HeavyGraph',
        annotations: Optional[Dict] = None
    ) -> None:
        assert obj_type == "send"
        super().__init__(obj_type, args, graph, annotations=annotations)

    def reduce(self) -> Optional[tuple]:
        if self.has_inlet_connection_format("c"):
            ir_args = dict(self.args)
            ir_args["hash"] = f"0x{HeavyLangObject.get_hash(ir_args['name']):X}"
            x = HIrSend("__send", ir_args, annotations=self.annotations)
            return ({x}, self.get_connection_move_list(x))

        elif self.has_inlet_connection_format("f"):
            # this case should be handled already in HeavyGraph.remap_send_receive()
            # clean up just in case
            return (set(), [])

        elif self.has_inlet_connection_format("_"):
            self.add_warning("send~ object doesn't have any inputs, this one is being removed.")
            return (set(), [])

        elif self.has_inlet_connection_format("m"):
            self.add_error(
                "Inlet can support either control or signal connections, "
                "but not both at the same time.")
            return None

        else:
            fmt = self._get_connection_format(self.inlet_connections)
            self.add_error(f"Unknown inlet configuration: {fmt}")
            return None



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
# Copyright (C) 2023 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from collections import defaultdict
from typing import Optional, Union, Dict, List

from .HeavyException import HeavyException
from .HeavyLangObject import HeavyLangObject


class LocalVars:
    """ A set of scoped objects.
    """

    def __init__(self, stdlib_dir: str = "./") -> None:
        # a dictionary of name-registered objects
        # the data structure is a list map
        # key is the name under which the object is registered, value is a
        # list of all objects who are registered with that name
        self.__REGISTERED_OBJ_DICT: Dict = defaultdict(list)

        # the list of globally declared paths
        self.declared_paths = [stdlib_dir]  # initialise with the standard library directory

    def find_path_for_abstraction(self, name: str) -> Optional[str]:
        # the file name based on the abstraction name
        file_name = f"{name}.hv.json"

        # iterate in order through the declared paths in order to find the file
        for d in self.declared_paths:
            file_path = os.path.join(d, file_name)
            if os.path.exists(file_path):
                return file_path  # if a matching abstraction is found, return the path
        return None  # otherwise return None

    def add_import_paths(self, path_list: List) -> None:
        """ Add import paths. Paths are expanded and made as explicit as possible.
        """
        self.declared_paths.extend([os.path.abspath(os.path.expanduser(p)) for p in path_list])

    def register_object(self, obj: HeavyLangObject, name: str, static: bool = False, unique: bool = False) -> None:
        """ Registers a named object.
        """

        objs = self.get_objects_for_name(name, obj.type)
        if (unique and len(objs) > 0) or (static and len(objs) > 1):
            # if there is already a registered object of this type and the new
            # object is not declared as static, throw an error
            raise HeavyException(f"Object {obj} with name \"{name}\" already exists, "
                                 "and the new object is static or unique.")
        elif len(objs) == 1 and static:
            pass  # the static object has already been registered, move on
        else:
            self.__REGISTERED_OBJ_DICT[name].append(obj)

    def unregister_object(self, obj: HeavyLangObject, name: str) -> None:
        """ Unregisters a named object.
        """

        self.__REGISTERED_OBJ_DICT[name].remove(obj)

    def get_objects_for_name(self, name: str, obj_types: Union[str, List]) -> List:
        """ Returns a list of objects registered under a name in this scope.
        """

        obj_types = obj_types if isinstance(obj_types, list) else [obj_types]
        return [o for o in self.__REGISTERED_OBJ_DICT[name] if o.type in obj_types]

    def get_registered_objects_for_type(self, obj_type: str) -> Dict:
        """ Returns a list-dictionary for all objects of the given type, indexed by name.
        """

        d = defaultdict(list)
        for k, v in self.__REGISTERED_OBJ_DICT.items():
            x = [o for o in v if o.type == obj_type]
            if len(x) > 0:  # only return names that actually have associated objects
                d[k].extend(x)
        return d
